#  Copyright 2022 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Union, Iterator
from urllib.parse import urlparse

import pandas as pd

from rikai.spark.sql.codegen.base import codegen_from_spec
from rikai.spark.sql.codegen.fs import FileSystemRegistry
from rikai.spark.sql.model import ModelSpec


def _make_model_spec(raw_spec: "ModelSpec") -> ModelSpec:
    uri = raw_spec["uri"]
    parsed = urlparse(uri)
    scheme = parsed.scheme
    if scheme == "file":
        reg = FileSystemRegistry()
    elif scheme == "mlflow":
        from rikai.spark.sql.codegen.mlflow_registry import MlflowRegistry

        reg = MlflowRegistry()
    elif scheme == "torchhub":
        from rikai.experimental.torchhub.torchhub_registry import (
            TorchHubRegistry,
        )

        reg = TorchHubRegistry()
    elif scheme == "tfhub":
        from rikai.experimental.tfhub.tfhub_registry import TFHubRegistry

        reg = TFHubRegistry()
    else:
        raise ValueError(f"Unknown scheme: {scheme}")
    return reg.make_model_spec(raw_spec)


def _apply_model_func(func, inputs):
    results_list = [results for results in func(inputs)]
    return results_list


def _func_from_spec(spec: ModelSpec):
    codegen = codegen_from_spec(spec)
    return codegen.generate_inference_func(spec)


def apply_model_spec(
    spec: "ModelSpec",
    inputs: Union[Iterator[pd.Series], Iterator[pd.DataFrame]],
) -> Iterator[pd.Series]:
    if not isinstance(spec, ModelSpec):
        spec = _make_model_spec(spec)
    func = _func_from_spec(spec)
    return _apply_model_func(func, inputs)



#  Copyright 2022 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""SSD: Single Shot MultiBox Detector

https://arxiv.org/abs/1512.02325
"""


from typing import Any, Callable, Dict

from rikai.tensorflow.models import TensorflowModelType
from rikai.types.geometry import Box2d

HUB_URL = "https://tfhub.dev/tensorflow/ssd_mobilenet_v2/2"


class SSDModelType(TensorflowModelType):
    def schema(self) -> str:
        return (
            "array<struct<detection_boxes:box2d,"
            "detection_scores:float, detection_classes:int>>"
        )

    def transform(self) -> Callable:
        return None

    def predict(self, images, *args, **kwargs) -> Any:
        assert (
            self.model is not None
        ), "model has not been initialized via load_model"
        batch = self.model(images)

        results = []
        for boxes, classes, scores in zip(
            batch["detection_boxes"].numpy(),
            batch["detection_classes"].numpy(),
            batch["detection_scores"].numpy(),
        ):
            predict_result = []
            for box, label_class, score in zip(boxes, classes, scores):
                predict_result.append(
                    {
                        "detection_boxes": Box2d(*box),
                        "detection_classes": int(label_class),
                        "detection_scores": float(score),
                    }
                )
            results.append(predict_result)
        return results


MODEL_TYPE = SSDModelType()



#  Copyright (c) 2021 Rikai Authors
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Custom Mlflow model logger to make sure models have the right logging for
Rikai SQL ML
"""

import os
import tempfile
import warnings
from typing import Any, Optional

CONF_MLFLOW_TRACKING_URI = "spark.rikai.sql.ml.registry.mlflow.tracking_uri"
CONF_MLFLOW_OUTPUT_SCHEMA = "rikai.output.schema"
CONF_MLFLOW_SPEC_VERSION = "rikai.spec.version"
CONF_MLFLOW_PRE_PROCESSING = "rikai.transforms.pre"
CONF_MLFLOW_POST_PROCESSING = "rikai.transforms.post"
CONF_MLFLOW_MODEL_FLAVOR = "rikai.model.flavor"
CONF_MLFLOW_MODEL_TYPE = "rikai.model.type"
CONF_MLFLOW_ARTIFACT_PATH = "rikai.model.artifact_path"


class MlflowLogger:
    """
    An alternative model logger for use during training instead of the vanilla
    mlflow logger.

    """

    _CURRENT_MODEL_SPEC_VERSION = "1.0"

    def __init__(self, flavor: str):
        """
        Parameters
        ----------
        flavor: str
            The model flavor
        """
        self.spec_version = MlflowLogger._CURRENT_MODEL_SPEC_VERSION
        self.flavor = flavor

    def _log_tensorflow_model(
        self,
        model: Any,
        artifact_path: str,
        registered_model_name: str,
        **kwargs,
    ):
        import tensorflow as tf

        try:
            import mlflow
        except ImportError as e:
            raise ImportError(
                "Couldn't import mlflow. Please make sure to "
                "`pip install mlflow` explicitly or install "
                "the correct extras like `pip install rikai[mlflow]`"
            ) from e

        with tempfile.TemporaryDirectory() as tmp_dir:

            tf.saved_model.save(model, tmp_dir)

            mlflow.tensorflow.log_model(
                tf_saved_model_dir=tmp_dir,
                tf_meta_graph_tags=[tf.saved_model.SERVING],
                tf_signature_def_key=next(iter(model.signatures.keys())),
                artifact_path=artifact_path,
                registered_model_name=registered_model_name,
                **kwargs,
            )

    def log_model(
        self,
        model: Any,
        artifact_path: str,
        schema: Optional[str] = None,
        pre_processing: Optional[str] = None,
        post_processing: Optional[str] = None,
        registered_model_name: Optional[str] = None,
        customized_flavor: Optional[str] = None,
        model_type: Optional[str] = None,
        **kwargs,
    ):
        """Convenience function to log the model with tags needed by rikai.
        This should be called during training when the model is produced.

        Parameters
        ----------
        model: Any
            The model artifact object
        artifact_path: str
            The relative (to the run) artifact path
        schema: str
            Output schema (pyspark DataType)
        pre_processing: str, default None
            Full python module path of the pre-processing transforms
        post_processing: str, default None
            Full python module path of the post-processing transforms
        registered_model_name: str, default None
            Model name in the mlflow model registry
        kwargs: dict
            Passed to `mlflow.<flavor>.log_model`


        Examples
        --------

        .. code-block:: python

            import rikai.mlflow

            # Log PyTorch model
            with mlflow.start_run() as run:

                # Training loop
                # ...

                # Assume `model` is the trained model from the training loop
                rikai.mlflow.pytorch.log_model(model, "model",
                        schema=rikai.contrib.torch.transforms.fasterrcnn_resnet50_fpn.OUTPUT_SCHEMA,
                        pre_processing="rikai.contrib.torch.transforms.fasterrcnn_resnet50_fpn.pre_processing",
                        post_processing="rikai.contrib.torch.transforms.fasterrcnn_resnet50_fpn.post_processing",
                        registered_model_name="MyPytorchModel")


        For more details see `mlflow docs <https://www.mlflow.org/docs/latest/python_api/mlflow.pytorch.html#mlflow.pytorch.log_model>`_.
        """  # noqa E501

        try:
            import mlflow
            from mlflow.tracking import MlflowClient
        except ImportError as e:
            raise ImportError(
                "Couldn't import mlflow. Please make sure to "
                "`pip install mlflow` explicitly or install "
                "the correct extras like `pip install rikai[mlflow]`"
            ) from e

        # no need to set the tracking uri here since this is intended to be
        # called inside the training loop within mlflow.start_run
        if self.flavor == "tensorflow":
            self._log_tensorflow_model(
                model,
                artifact_path,
                registered_model_name=registered_model_name,
                **kwargs,
            )
        else:
            getattr(mlflow, self.flavor).log_model(
                model,
                artifact_path,
                registered_model_name=registered_model_name,
                **kwargs,
            )

        tags = {
            CONF_MLFLOW_SPEC_VERSION: MlflowLogger._CURRENT_MODEL_SPEC_VERSION,
            CONF_MLFLOW_MODEL_FLAVOR: customized_flavor
            if customized_flavor
            else self.flavor,
            CONF_MLFLOW_MODEL_TYPE: model_type,
            CONF_MLFLOW_OUTPUT_SCHEMA: schema,
            CONF_MLFLOW_PRE_PROCESSING: pre_processing,
            CONF_MLFLOW_POST_PROCESSING: post_processing,
            CONF_MLFLOW_ARTIFACT_PATH: artifact_path,
        }
        for k in (
            CONF_MLFLOW_PRE_PROCESSING,
            CONF_MLFLOW_POST_PROCESSING,
            CONF_MLFLOW_MODEL_TYPE,
            CONF_MLFLOW_OUTPUT_SCHEMA,
        ):
            if not tags[k]:
                del tags[k]
                warnings.warn(
                    f"value of {k} is None or empty and "
                    "will not be populated to MLflow"
                )
        mlflow.set_tags(tags)
        if registered_model_name is not None:
            # if we're creating a model registry entry,
            # we also want to set the tags on the model version
            # and model to enable search
            c = MlflowClient()
            # mlflow log_model does not return the version (wtf)
            all_versions = c.get_latest_versions(
                registered_model_name, stages=["production", "staging", "None"]
            )
            current_version = None
            run_id = mlflow.active_run().info.run_id
            for v in all_versions:
                if v.run_id == run_id:
                    current_version = v
                    break
            if current_version is None:
                raise ValueError(
                    "No model version found matching runid: {}".format(run_id)
                )
            for key, value in tags.items():
                c.set_registered_model_tag(registered_model_name, key, value)
                c.set_model_version_tag(
                    registered_model_name, current_version.version, key, value
                )


KNOWN_FLAVORS = ["pytorch", "sklearn", "tensorflow"]
for flavor in KNOWN_FLAVORS:
    globals()[flavor] = MlflowLogger(flavor)



#  Copyright 2021 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
from pathlib import Path
from typing import Optional, Union
from urllib.parse import urlparse

import yaml

from rikai.io import open_uri
from rikai.logging import logger
from rikai.spark.sql.codegen.base import ModelSpec, Registry, udf_from_spec
from rikai.spark.sql.exceptions import SpecError

__all__ = ["FileSystemRegistry"]


class FileModelSpec(ModelSpec):
    """Model Spec.

    Parameters
    ----------
    spec_uri : str or Path
        Spec file URI
    options : Dict[str, Any], optional
        Additionally options. If the same option exists in spec already,
        it will be overridden.
    validate : bool, default True.
        Validate the spec during construction. Default ``True``.
    """

    def __init__(
        self,
        spec_uri: Union[str, Path],
        options: Optional[dict] = None,
        validate: bool = True,
    ):
        with open_uri(spec_uri) as fobj:
            spec = yaml.load(fobj, Loader=yaml.FullLoader)
        self.base_dir = os.path.dirname(spec_uri)
        spec.setdefault("options", {})
        if options:
            spec["options"].update(options)
        super().__init__(spec, validate=validate)

    def load_model(self):
        if self.flavor == "pytorch":
            from rikai.spark.sql.codegen.pytorch import load_model_from_uri

            return load_model_from_uri(self.model_uri)
        elif self.flavor == "tensorflow":
            from rikai.spark.sql.codegen.tensorflow import load_model_from_uri

            return load_model_from_uri(self.model_uri)
        else:
            raise SpecError("Unsupported flavor {}".format(self.flavor))

    @property
    def model_uri(self):
        """Absolute model URI."""
        origin_uri = super().model_uri
        parsed = urlparse(origin_uri)
        if parsed.scheme or os.path.isabs(origin_uri):
            return origin_uri
        return os.path.join(self.base_dir, origin_uri)


class FileSystemRegistry(Registry):
    """FileSystem-based Model Registry"""

    def __repr__(self):
        return "FileSystemRegistry"

    def make_model_spec(self, raw_spec: dict):
        uri = raw_spec.get("uri")
        options = raw_spec.get("options", {})
        spec = FileModelSpec(uri, options=options)
        return spec



#  Copyright (c) 2021 Rikai Authors
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""
Various s3 utilities for Rikai functionality. For example, libraries like
OpenCV cannot open s3 url's directly. Instead we make a presigned s3 url.
"""
import logging

try:
    import boto3
    from botocore.exceptions import ClientError
except ImportError as e:
    raise ImportError("Please `pip install rikai[aws]` to use s3 utils") from e


def create_presigned_url(bucket_name, object_name, expiration=3600):
    """Generate a presigned URL to share an S3 object

    Parameters
    ----------
    bucket_name: str
        S3 bucket name
    object_name: str
        S3 key path (everything after the bucket name)
    expiration: int
        Expiration in seconds
    """
    s3 = boto3.client("s3")
    try:
        return s3.generate_presigned_url(
            "get_object",
            Params={"Bucket": bucket_name, "Key": object_name},
            ExpiresIn=expiration,
        )
    except ClientError as e:
        logging.error(e)
        return None



#  Copyright 2022 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Inspect detail of Torchvision SSD model

https://pytorch.org/vision/stable/models.html#torchvision.models.detection.ssd300_vgg16

"""

from typing import Callable, Dict, List, Tuple

import torch
import torch.nn.functional as F
from torch import Tensor
from torchvision.models.detection.ssd import SSD
from torchvision.ops.boxes import batched_nms, clip_boxes_to_image

from rikai.contrib.torch.transforms.ssd import pre_processing
from rikai.types import Box2d

__all__ = [
    "SSDClassScoresExtractor",
    "class_scores_extractor_post_processing",
    "class_scores_extractor_pre_processing",
]


class SSDClassScoresExtractor(torch.nn.Module):
    """Extracts the scores (confidences) of each class for
    all the detected bounding box.

    Parameters
    ----------
    backend : :class:`torch.nn.Module`
        The trained SSD model
    topk_candidates : int, optional
        The number of top candidates (classes) returned per box.

    Returns
    -------
    [Dict[str, Tensor]]
        With the form of
        ``{"boxes": FloatTensor[N, 4], "labels": Int64Tensor[N, k], "scores": Tensor[N, k]}``
    """  # noqa: E501

    # SCHEMA to be used in Rikai SQL ML
    SCHEMA = (
        "array<struct<box:box2d, scores:array<float>, label_ids:array<int>>>"
    )

    def __init__(self, backend: SSD, topk_candidates: int = 2):
        super().__init__()
        if not isinstance(backend, SSD):
            raise ValueError("Only support Torchvision's SSD model.")
        self.backend = backend
        self.topk_candidates = topk_candidates

    def forward(
        self,
        images: List[torch.Tensor],
    ):
        if self.training:
            raise ValueError("This feature extractor only supports eval mode.")

        # Reuse the code in torchvision's SSD.forward(), but returns multiple
        # classes and scores for each box.
        # TODO(GH-517): abstract the confidence extraction code for multiple
        #       torchvision models.

        # get the original image sizes
        original_image_sizes: List[Tuple[int, int]] = []
        for img in images:
            val = img.shape[-2:]
            assert len(val) == 2
            original_image_sizes.append((val[0], val[1]))

        images, _ = self.backend.transform(images, None)

        # Get the features from the backbone
        features = self.backend.backbone(images.tensors)
        features = list(features.values())
        head_outputs = self.backend.head(features)
        pred_anchors = self.backend.anchor_generator(images, features)

        scores = head_outputs["cls_logits"]
        bbox_regression = head_outputs["bbox_regression"]
        pred_scores = F.softmax(scores, dim=-1)

        num_classes = pred_scores.size(-1)
        device = pred_scores.device

        detections: List[Dict[str, Tensor]] = []
        for boxes, scores, anchors, image_shape in zip(
            bbox_regression, pred_scores, pred_anchors, images.image_sizes
        ):
            boxes = self.backend.box_coder.decode_single(boxes, anchors)
            boxes = clip_boxes_to_image(boxes, image_shape)

            image_boxes = []
            image_scores = []
            image_labels = []
            image_all_scores = []
            for label in range(1, num_classes):
                score = scores[:, label]

                keep_idxs = score > self.backend.score_thresh
                score = score[keep_idxs]
                all_scores = scores[keep_idxs]
                box = boxes[keep_idxs]

                # keep only topk scoring predictions
                num_topk = min(self.backend.topk_candidates, score.size(0))
                score, idxs = score.topk(num_topk)
                box = box[idxs]
                all_scores = all_scores[idxs]

                image_boxes.append(box)
                image_scores.append(score)
                image_all_scores.append(all_scores)
                image_labels.append(
                    torch.full_like(
                        score,
                        fill_value=label,
                        dtype=torch.int64,
                        device=device,
                    )
                )

            image_boxes = torch.cat(image_boxes, dim=0)
            image_scores = torch.cat(image_scores, dim=0)
            image_labels = torch.cat(image_labels, dim=0)
            image_all_scores = torch.cat(image_all_scores, dim=0)

            # non-maximum suppression
            keep = batched_nms(
                image_boxes,
                image_scores,
                image_labels,
                self.backend.nms_thresh,
            )
            keep = keep[: self.backend.detections_per_img]
            top_scores, idx = image_all_scores[keep, 1:].topk(
                self.topk_candidates
            )
            detections.append(
                {
                    "boxes": image_boxes[keep],
                    "scores": top_scores,
                    "labels": idx + 1,
                }
            )
        detections = self.backend.transform.postprocess(
            detections, images.image_sizes, original_image_sizes
        )
        return detections


class_scores_extractor_pre_processing = pre_processing


def class_scores_extractor_post_processing(
    options: Dict[str, str]
) -> Callable:
    def post_process_func(batch):
        results = []
        for predicts in batch:
            predict_result = []
            for box, label, score in zip(
                predicts["boxes"].tolist(),
                predicts["labels"].tolist(),
                predicts["scores"].tolist(),
            ):
                predict_result.append(
                    {
                        "box": Box2d(*box),
                        "label_ids": label,
                        "scores": score,
                    }
                )

            results.append(predict_result)
        return results

    return post_process_func



#  Copyright 2020 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Parquet-based Rikai Dataset that supports automatically UDT conversions.
"""

# Standard Library
import importlib
from functools import partial
from pathlib import Path
from typing import Any, Dict, List, Mapping, Optional, Union

# Third Party
import numpy as np
import pandas as pd
import pyarrow.dataset as ds
import pyarrow.parquet as pq
from pyarrow import fs
from pyspark.ml.linalg import Matrix, Vector
from pyspark.sql import Row
from pyspark.sql.types import UserDefinedType

# Rikai
from rikai.io import open_input_stream
from rikai.logging import logger
from rikai.mixin import ToNumpy, ToPIL
from rikai.parquet.resolver import Resolver
from rikai.parquet.shuffler import RandomShuffler

__all__ = ["Dataset"]


class Dataset:
    """Rikai Dataset.

    :py:class:`Dataset` provides read access to a Rikai dataset. It supports

    - Read Rikai encoded dataset on the supported storage medias, i.e.,
      local filesystem, AWS S3 or Google GCS.
    - Automatically deserialize data into semantic user defined types (UDT).
    - Shuffle the dataset randomly through ``shuffle`` flag.
    - Distributed training by setting ``world_size`` and ``rank`` parameters.
      When enabled, parquet `row-group` level partition will be used to
      distribute data amount the workers.

    Parameters
    ----------
    query : str
        A SQL query "SELECT image, annotation FROM s3://foo/bar" or
        a dataset URI, i.e., "s3://foo/bar"
    columns : List[str], optional
        To read only given columns
    shuffle : bool, optional
        Set to True to shuffle the results.
    shuffler_capacity : int
        The size of the buffer to shuffle the examples. The size of buffer
        does not impact the distribution of possibility that an example
        is picked.
    seed : int, optional
        Random seed for shuffling process.
    world_size : int
        Total number of distributed workers
    rank : int
        The rank of this worker in all the distributed workers

    Notes
    -----
    - Typically user should not directly use this class. Instead, users are
      encouraged to use framework-native readers, for example, using
      :py:class:`rikai.pytorch.data.Dataset` in
      `Pytorch <https://pytorch.org/>`_
    """

    SPARK_PARQUET_ROW_METADATA = b"org.apache.spark.sql.parquet.row.metadata"

    _UDT_CACHE: Dict[str, UserDefinedType] = {}

    def __init__(  # pylint: disable=too-many-arguments
        self,
        query: Union[str, Path],
        columns: Optional[List[str]] = None,
        shuffle: bool = False,
        shuffler_capacity: int = 128,
        seed: Optional[int] = None,
        world_size: int = 1,
        rank: int = 0,
    ):
        self.uri = str(query)
        self.columns = columns
        self.shuffle = shuffle
        self.shuffler_capacity = shuffler_capacity
        self.seed = seed
        self.rank = rank
        self.world_size = world_size
        if self.world_size > 1:
            logger.info(
                "Running in distributed mode, world size=%s, rank=%s",
                world_size,
                rank,
            )

        # Provide deterministic order between distributed workers.
        self.files = sorted(Resolver.resolve(self.uri))
        if self.rank == 0:
            logger.info("Loading parquet files: %s", self.files)

        self.spark_row_metadata = Resolver.get_schema(self.uri)

    def __repr__(self) -> str:
        return "Dataset(uri={}, columns={}, shuffle={})".format(
            self.uri, self.columns if self.columns else "[*]", self.shuffle
        )

    @classmethod
    def _find_udt(cls, pyclass: str) -> UserDefinedType:
        """Find UDT class specified by the python class path."""
        if pyclass in cls._UDT_CACHE:
            return cls._UDT_CACHE[pyclass]

        class_path = pyclass.split(".")
        module_name = ".".join(class_path[:-1])
        class_name = class_path[-1]
        try:
            udt_class = getattr(
                importlib.import_module(module_name), class_name
            )
        except ImportError as err:
            raise ImportError(
                f"Could not import user defind type {pyclass}"
            ) from err
        cls._UDT_CACHE[pyclass] = udt_class()
        return cls._UDT_CACHE[pyclass]

    def _convert(
        self, raw_row: Dict[str, Any], schema: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Convert Spark UDT to native rikai or numpy types.

        Parameters
        ----------
        raw_row : Dict[str, Any]
        schema : Dict[str, Any]
            Spark schema in the JSON format
        """
        assert schema["type"] == "struct"
        converted = {}
        for field in schema["fields"]:
            name = field["name"]
            if name not in raw_row:
                # This column is not selected, skip
                continue
            field_type = field["type"]
            if not isinstance(field_type, dict):
                converted[name] = raw_row[name]
            elif field_type["type"] == "udt":
                udt = self._find_udt(field_type["pyClass"])
                converted[name] = _convert_udt_value(raw_row[name], udt)
            elif field_type["type"] == "array" and isinstance(
                field_type["elementType"], dict
            ):
                converted[name] = self._convert_array(
                    raw_row[name], field_type["elementType"]
                )
            elif field_type["type"] == "struct":
                converted[name] = {
                    f["name"]: self._convert(raw_row[f["name"]], f["type"])
                    for f in field_type["type"]["fields"]
                }
            else:
                converted[name] = raw_row[name]

        return converted

    def __iter__(self):
        shuffler = RandomShuffler(
            self.shuffler_capacity if self.shuffle else 1, self.seed
        )
        group_count = -1
        for file_uri in self.files:
            with open_input_stream(file_uri) as fobj:
                parquet = pq.ParquetFile(fobj)
                for group_idx in range(parquet.num_row_groups):
                    # A simple form of row-group level bucketing without
                    # memory overhead.
                    # Pros:
                    #  - It requires zero communication to initialize the
                    #    distributed policy
                    #  - It uses little memory and no startup overhead, i.e.
                    #    collecting row groups.
                    # Cons:
                    #   The drawback would be if the world size is much larger
                    #   than the average number of row groups. As a result,
                    #   many of the file open operations would be wasted.
                    group_count += 1
                    if group_count % self.world_size != self.rank:
                        continue
                    row_group = parquet.read_row_group(
                        group_idx, columns=self.columns
                    )
                    for (
                        batch
                    ) in row_group.to_batches():  # type: pyarrow.RecordBatch
                        # TODO: read batches not using pandas
                        for _, row in batch.to_pandas().iterrows():
                            shuffler.append(row)
                            # Maintain the shuffler buffer around its capacity.

                            while shuffler.full():
                                yield self._convert(
                                    shuffler.pop().to_dict(),
                                    self.spark_row_metadata,
                                )
        while shuffler:
            yield self._convert(
                shuffler.pop().to_dict(), self.spark_row_metadata
            )

    def to_pandas(self, limit=None):
        """Create a pandas dataframe from the parquet data in this Dataset

        Parameters
        ----------
        limit: int, default None
            The max number of rows to retrieve. If none, 0, or negative
            then retrieve all rows
        """
        filesystem, path = fs.FileSystem.from_uri(self.uri)
        dataset = ds.dataset(path, filesystem=filesystem, format="parquet")
        if limit is None or limit <= 0:
            raw_df = dataset.to_table(columns=self.columns).to_pandas()
        else:
            raw_df = dataset.head(limit, columns=self.columns).to_pandas()
        types = {
            f["name"]: f["type"] for f in self.spark_row_metadata["fields"]
        }
        return pd.DataFrame(
            {
                name: self._convert_col(col, types.get(name))
                for name, col in raw_df.iteritems()
            }
        )

    def _convert_col(self, col: pd.Series, field_type) -> pd.Series:
        if field_type is None:
            return col
        if not isinstance(field_type, dict):
            return col
        elif field_type["type"] == "udt":
            udt = self._find_udt(field_type["pyClass"])
            return col.apply(partial(_convert_udt_value, udt=udt))
        elif field_type["type"] == "struct":
            return col.apply(lambda d: self._convert(d, field_type))
        elif field_type["type"] == "array":
            return col.apply(
                lambda d: self._convert_array(d, field_type["elementType"])
            )
        else:
            return col

    def _convert_array(self, arr, element_type):
        return [self._convert(x, element_type) for x in arr]


def _convert_udt_value(value, udt):
    if isinstance(value, dict):
        row = Row(**value)
    else:
        row = Row(value)
    converted_value = udt.deserialize(row)
    if isinstance(converted_value, (Vector, Matrix)):
        converted_value = converted_value.toArray()
    return converted_value


def convert_tensor(row, use_pil: bool = False):
    """
    Convert a parquet row into tensors.

    If use_pil is set to True, this method returns a PIL image instead,
    and relies on the customer code to convert PIL image to tensors.
    """
    if use_pil and isinstance(row, ToPIL):
        return row.to_pil()
    elif isinstance(row, ToNumpy):
        return row.to_numpy()
    elif not isinstance(row, (Mapping, pd.Series)):
        # Primitive values
        return row

    tensors = {}
    for key, value in row.items():
        if isinstance(value, dict):
            tensors[key] = convert_tensor(value)
        elif isinstance(value, (list, tuple)):
            tensors[key] = np.array([convert_tensor(elem) for elem in value])
        elif use_pil and isinstance(value, ToPIL):
            tensors[key] = value.to_pil()
        elif isinstance(value, ToNumpy):
            tensors[key] = value.to_numpy()
        else:
            tensors[key] = value
    return tensors



#  Copyright 2022 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Pytorch Dataset and DataLoader"""

import os
import uuid
from pathlib import Path
from typing import Callable, List, Union

# Third Party
import torch
from torch.utils.data import IterableDataset

# Rikai
import rikai.parquet
from rikai.pytorch.transforms import RikaiToTensor
from rikai.spark.utils import df_to_rikai

__all__ = ["Dataset"]


class Dataset(IterableDataset):
    """Rikai Pytorch Dataset.

    A :py:class:`torch.utils.data.IterableDataset` that reads
    Rikai data format. This :py:class:`Dataset` works with
    `multi-process data loading`_ using :py:class:`torch.utils.data.DataLoader`.

    Parameters
    ----------
    data_ref : str, Path, pyspark.sql.DataFrame
        URI to the data files or the dataframe
    columns : list of str, optional
        An optional list of column to load from parquet files.
    transform: Callable, default instance of RikaiToTensor
        Apply row level transformation before yielding each sample

    Note
    ----

    Up to ``pytorch==1.7``, :py:class:`~torch.utils.data.IterableDataset`
    does not work with :py:class:`torch.utils.data.Sampler` with
    :py:class:`torch.utils.data.DataLoader`.

    Use :py:class:`torch.utils.data.BufferedShuffleDataset` (torch>=1.8)
    with the Rikai dataset for randomness.

    Example
    -------

    >>> from rikai.pytorch.data import Dataset
    >>> from torch.utils.data import DataLoader
    >>>
    >>> dataset = Dataset("dataset", columns=["image", "label_id"])
    >>> # dataset = BufferedShuffleDataset(dataset)
    >>> loader = DataLoader(dataset, num_workers=8)

    .. _multi-process data loading: https://pytorch.org/docs/master/data.html#single-and-multi-process-data-loading
    """  # noqa: E501

    def __init__(
        self,
        data_ref: Union[str, Path, "pyspark.sql.DataFrame"],
        columns: List[str] = None,
        transform: Callable = RikaiToTensor(),
    ):
        super().__init__()
        self.data_ref = data_ref
        self.columns = columns
        self._transform = transform

    def __repr__(self) -> str:
        return f"Dataset(torch, {self.data_ref}, columns={self.columns})"

    def __iter__(self):
        rank = 0
        world_size = 1

        worker_info = torch.utils.data.get_worker_info()
        if worker_info is not None:
            rank = worker_info.id
            world_size = worker_info.num_workers

        uri = _maybe_cache_df(self.data_ref)
        for row in rikai.parquet.Dataset(
            uri,
            columns=self.columns,
            world_size=world_size,
            rank=rank,
        ):
            yield self._transform(row)


def _maybe_cache_df(
    data_ref: Union[str, Path, "pyspark.sql.DataFrame"]
) -> str:
    """
    If the given dataset_ref is a str/Path, then just return a str ref.
    If it's a pyspark DataFrame then cache the DataFrame as parquet and
    return the cache uri

    Parameters
    ----------
    data_ref: str, Path, or pyspark DataFrame
        Either a uri to parquet or a DataFrame that will be cached as parquet

    Returns
    -------
    uri: str
        Either the input parquet uri or the cached uri for a DataFrame
    """
    if isinstance(data_ref, (str, Path)):
        return str(data_ref)
    else:
        try:
            from pyspark.sql import DataFrame
        except ImportError:
            raise ImportError(
                "Cannot create rikai pytorch dataset from "
                "Spark DataFrame without pyspark installed."
            )
        if isinstance(data_ref, DataFrame):
            cache_uri = _get_cache_uri(data_ref)
            df_to_rikai(data_ref, cache_uri)
        else:
            raise TypeError(
                (
                    "dataset_ref must be a str, Path, or DataFrame and "
                    "not a {}."
                ).format(type(data_ref))
            )
        return cache_uri


def _get_cache_uri(df: "pyspark.sql.DataFrame") -> str:
    """
    TODO create a deterministic unique uri from the df for sharing

    Parameters
    ----------
    df: DataFrame
        The cache uri will be generated for the given DataFrame

    Returns
    -------
    cache_uri: str
        The uri to write the DataFrame to
    """
    cache_root_uri = rikai.options.rikai.cache_uri
    if not cache_root_uri:
        raise ValueError(
            "Could not retrieve rikai cache_uri from either "
            "spark or rikai configurations."
        )
    return os.path.join(cache_root_uri, str(uuid.uuid4()))



#  Copyright (c) 2021 Rikai Authors
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import re

import rikai
from rikai.__version__ import version
from rikai.conf import CONF_PARQUET_BLOCK_SIZE


def df_to_rikai(df: "pyspark.sql.DataFrame", uri: str):
    (
        df.write.format("rikai")
        .option(CONF_PARQUET_BLOCK_SIZE, rikai.options.parquet.block.size)
        .save(uri)
    )


def get_default_jar_version(use_snapshot=True):
    """
    Make it easier to reference the jar version in notebooks and conftest.

    Parameters
    ----------
    use_snapshot: bool, default True
        If True then map `*dev0` versions to `-SNAPSHOT`
    """
    pattern = re.compile(r"([\d]+.[\d]+.[\d]+)")
    match = re.search(pattern, version)
    if not match:
        raise ValueError("Ill-formed version string {}".format(version))
    match_str = match.group(1)
    if use_snapshot and (len(match_str) < len(version)):
        return match_str + "-SNAPSHOT"
    return match_str


def init_spark_session(
    conf: dict = None, app_name="rikai", rikai_version=None
):
    from pyspark.sql import SparkSession

    # Avoid reused session polluting configs
    active_session = SparkSession.getActiveSession()
    if active_session and conf:
        for k, v in conf.items():
            if str(active_session.conf.get(k)) != str(v):
                print(
                    f"active session: want {v} for {k}"
                    f" but got {active_session.conf.get(k)},"
                    f" will restart session"
                )
                active_session.stop()
                break

    if not rikai_version:
        rikai_version = get_default_jar_version(use_snapshot=True)
    builder = (
        SparkSession.builder.appName(app_name)
        .config(
            "spark.jars.packages",
            ",".join(
                [
                    "ai.eto:rikai_2.12:{}".format(rikai_version),
                ]
            ),
        )
        .config(
            "spark.sql.extensions",
            "ai.eto.rikai.sql.spark.RikaiSparkSessionExtensions",
        )
        .config(
            "spark.driver.extraJavaOptions",
            "-Dio.netty.tryReflectionSetAccessible=true",
        )
        .config(
            "spark.executor.extraJavaOptions",
            "-Dio.netty.tryReflectionSetAccessible=true",
        )
    )
    conf = conf or {}
    for k, v in conf.items():
        builder = builder.config(k, v)
    session = builder.master("local[2]").getOrCreate()
    return session



#  Copyright (c) 2021 Rikai Authors
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""
Useful functions and features for working with video data
"""

from typing import Optional, Union

# Standard library
from urllib.parse import urlparse

# Third Party
from pyspark.sql.functions import udf
from pyspark.sql.types import (
    ArrayType,
    FloatType,
    IntegerType,
    StringType,
    StructField,
    StructType,
)

# Rikai
from rikai.types.video import (
    Segment,
    SingleFrameSampler,
    VideoStream,
    YouTubeVideo,
)

__all__ = ["scene_detect", "SCENE_LIST_SCHEMA"]


SCENE_LIST_SCHEMA = ArrayType(
    StructType(
        [
            StructField(
                "start",
                StructType(
                    [
                        StructField("frame_num", IntegerType(), True),
                        StructField("frame_pos_sec", FloatType(), True),
                    ]
                ),
                True,
            ),
            StructField(
                "end",
                StructType(
                    [
                        StructField("frame_num", IntegerType(), True),
                        StructField("frame_pos_sec", FloatType(), True),
                    ]
                ),
                True,
            ),
        ]
    )
)


@udf(returnType=SCENE_LIST_SCHEMA)
def scene_detect(
    videos: Union[str, VideoStream, list], threshold=30.0, raise_on_error=True
):
    """Use `pyscenedetect <https://pyscenedetect.readthedocs.io/en/latest/>`_
    to split scenes in a video.

    Parameters
    ----------
    videos: str, VideoStream, or list<str>
        The video to split. If a list is given, assume they are segments of a
        longer video
    threshold: float, default 30.0
        Detection threshold (i.e., ContentDetector(threshold=threshold))
    raise_on_error: bool, default True
        Set to False to suppress exceptions during the detection run (note that
        import errors will still be raised)
    """
    try:
        from scenedetect.detectors import ContentDetector
        from scenedetect.scene_manager import SceneManager
        from scenedetect.video_manager import VideoManager
    except ImportError as e:
        raise ImportError(
            (
                "Please make sure you have pyscenedetect installed "
                "via `pip install rikai[video]`"
            )
        )
    video_manager = None
    try:
        if isinstance(videos, VideoStream):
            video_path = [videos.uri]
        elif isinstance(videos, list):
            video_path = videos
        else:
            video_path = [videos]

        video_manager = VideoManager(normalize_uri(video_path))
        scene_manager = SceneManager()  # TODO add StatsManager later
        scene_manager.add_detector(ContentDetector(threshold=threshold))
        video_manager.set_downscale_factor()
        video_manager.start()
        scene_manager.detect_scenes(frame_source=video_manager)
        scene_list = scene_manager.get_scene_list()
        return [
            {
                "start": {
                    "frame_num": s[0].frame_num,
                    "frame_pos_sec": s[0].get_seconds(),
                },
                "end": {
                    "frame_num": s[1].frame_num,
                    "frame_pos_sec": s[1].get_seconds(),
                },
            }
            for s in scene_list
        ]
    except Exception as e:
        if raise_on_error:
            raise e
        else:
            return None
    finally:
        if video_manager is not None:
            video_manager.release()


def normalize_uri(uri: Union[str, list]):
    if isinstance(uri, str):
        parsed = urlparse(uri)
        if parsed.scheme == "s3":
            from rikai.contrib.s3 import create_signed_url

            return create_signed_url(parsed.netloc, parsed.path)
        return uri
    else:
        return [normalize_uri(u) for u in uri]



#  Copyright 2021 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
from typing import Iterator

import pandas as pd
import torch
from pyspark.serializers import CloudPickleSerializer
from pyspark.sql.functions import pandas_udf
from pyspark.sql.types import BinaryType
from torch.utils.data import DataLoader

from rikai.io import open_uri
from rikai.pytorch.pandas import PandasDataset
from rikai.spark.sql.model import AnonymousModelType, ModelSpec

DEFAULT_NUM_WORKERS = 8
DEFAULT_BATCH_SIZE = 4

_pickler = CloudPickleSerializer()


def move_tensor_to_device(data, device):
    if isinstance(data, torch.Tensor):
        return data.to(device)
    elif isinstance(data, list):
        return [move_tensor_to_device(elem, device) for elem in data]
    # Do nothing
    return data


def _generate(payload: ModelSpec, is_udf: bool = True):
    """Construct a UDF to run pytorch model.

    Parameters
    ----------
    payload : ModelSpec
        the model specifications object

    Returns
    -------
    A Spark Pandas UDF.
    """
    model = payload.model_type
    default_device = "gpu" if torch.cuda.is_available() else "cpu"
    options = payload.options
    use_gpu = options.get("device", default_device) == "gpu"
    num_workers = int(
        options.get("num_workers", min(os.cpu_count(), DEFAULT_NUM_WORKERS))
    )
    batch_size = int(options.get("batch_size", DEFAULT_BATCH_SIZE))

    return_type = Iterator[pd.Series]

    def torch_inference_udf(
        iter: Iterator[pd.DataFrame],
    ) -> return_type:
        device = torch.device("cuda" if use_gpu else "cpu")
        model.load_model(payload, device=device)
        if isinstance(model, AnonymousModelType):
            # We will remove them after AnonymousModelType deprecation
            model.model.eval()
            model.model.to(device)

        try:
            with torch.no_grad():
                for series in iter:
                    dataset = PandasDataset(
                        series,
                        transform=model.transform(),
                        unpickle=is_udf,
                        use_pil=True,
                    )
                    results = []
                    for batch in DataLoader(
                        dataset,
                        batch_size=batch_size,
                        num_workers=num_workers,
                    ):
                        batch = move_tensor_to_device(batch, device)
                        predictions = model(batch)
                        bin_predictions = [
                            _pickler.dumps(p) if is_udf else p
                            for p in predictions
                        ]
                        results.extend(bin_predictions)
                    yield pd.Series(results)
        finally:
            if use_gpu:
                model.release()

    if is_udf:
        return pandas_udf(torch_inference_udf, returnType=BinaryType())
    else:
        return torch_inference_udf


def generate_inference_func(payload: ModelSpec):
    return _generate(payload, False)


def generate_udf(payload: ModelSpec):
    return _generate(payload, True)


def load_model_from_uri(uri: str):
    with open_uri(uri) as fobj:
        return torch.load(fobj)



#  Copyright 2021 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Convert Coco dataset into Rikai format.

https://cocodataset.org/#home
"""

from __future__ import annotations

import json
import os
from itertools import islice
from pathlib import Path
from typing import Optional, Union

from pyspark.sql import DataFrame, SparkSession
from pyspark.sql.functions import col, collect_list, lit, struct, udf
from pyspark.sql.types import (
    ArrayType,
    DoubleType,
    FloatType,
    IntegerType,
    LongType,
    StringType,
    StructField,
    StructType,
)

try:
    from pycocotools.coco import COCO
except ImportError as exc:
    raise ImportError("Please install pycocotools") from exc

from rikai.io import open_uri
from rikai.spark.functions import image_copy
from rikai.spark.types import Box2dType, ImageType, MaskType
from rikai.types import Box2d, Image, Mask

__all__ = ["convert"]

ANNOTATION_TYPE = StructType(
    [
        StructField("image_id", IntegerType()),
        StructField("area", DoubleType()),
        StructField("label_id", IntegerType()),
        StructField("ann_id", LongType()),
        StructField("bbox", Box2dType()),
        StructField("segmentation", MaskType()),
        StructField("supercategory", StringType()),
        StructField("label", StringType()),
    ]
)


def load_categories(annotation_file):
    """Load categories to be global categories."""
    with open(annotation_file) as fobj:
        ann = json.load(fobj)
        return {c["id"]: c for c in ann["categories"]}


@udf(returnType=ArrayType(ANNOTATION_TYPE))
def rescale_bbox(annotations, height, width):
    """Rescale bounding box to relative scale [0, 1] to the image."""
    ret = []
    for ann in annotations:
        ann = ann.asDict()
        ann["bbox"] = ann["bbox"] / (width, height)
        ret.append(ann)
    return ret


def convert_instance(
    spark: SparkSession,
    annotation_json: Union[str, Path],
    image_dir: Union[str, Path],
) -> DataFrame:

    with open_uri(annotation_json) as fobj:
        data = json.load(fobj)
    images_map = {img["id"]: img for img in data["images"]}
    total_images = len(images_map)

    categories_map = {c["id"]: c for c in data["categories"]}

    images = [
        {
            "id": img["id"],
            "date_captured": img["date_captured"],
            "width": img["width"],
            "height": img["height"],
            "file_name": img["file_name"],
            "image": Image(Path(image_dir) / img["file_name"]),
        }
        for img in data["images"]
    ]
    image_df = spark.createDataFrame(
        images,
        schema=StructType(
            [
                StructField("id", IntegerType()),
                StructField("date_captured", StringType()),
                StructField("width", IntegerType()),
                StructField("height", IntegerType()),
                StructField("file_name", StringType()),
                StructField("image", ImageType()),
            ]
        ),
    ).repartition(max(1, total_images // 500))

    annotations_df = spark.createDataFrame(
        [
            {
                "image_id": ann["image_id"],
                "area": float(ann["area"]),
                "label_id": ann["category_id"],
                "ann_id": ann["id"],
                "bbox": Box2d.from_top_left(*ann["bbox"]),
                "segmentation": Mask.from_coco_rle(
                    ann["segmentation"]["counts"],
                    height=ann["segmentation"]["size"][0],
                    width=ann["segmentation"]["size"][1],
                )
                if ann["iscrowd"]
                else Mask.from_polygon(
                    ann["segmentation"],
                    height=images_map[ann["image_id"]]["height"],
                    width=images_map[ann["image_id"]]["width"],
                ),
                "label": categories_map[ann["category_id"]]["name"],
                "supercategory": categories_map[ann["category_id"]][
                    "supercategory"
                ],
            }
            for ann in data["annotations"]
        ],
        schema=ANNOTATION_TYPE,
    )

    annotations_per_image = annotations_df.groupBy("image_id").agg(
        collect_list(struct(annotations_df.columns)).alias("raw_annotations")
    )

    # Join annotations first then load images into the dataset
    images_meta_df = (
        image_df.join(
            annotations_per_image,
            image_df.id == annotations_per_image.image_id,
        )
        .withColumn(
            "annotations", rescale_bbox("raw_annotations", "height", "width")
        )
        .drop("id", "raw_annotations")
    )
    return images_meta_df


def convert(
    spark: SparkSession,
    dataset_root: str,
    limit: int = 0,
    asset_dir: Optional[str] = None,
) -> DataFrame:
    """Convert a Coco Dataset into Rikai dataset.

    This function expects the COCO datasets are stored in directory with the
    following structure:

    - dataset
        - annotations
          - captions_train2017.json
          - instances_train2017.json
          - ...
        - train2017
        - val2017
        - test2017

    Parameters
    ----------
    spark : SparkSession
        A live spark session
    dataset_root : str
        The directory of dataset
    limit : int, optional
        The number of images of each split to be converted.
    asset_dir : str, optional
        The asset directory to store images, can be a s3 directory.

    Return
    ------
    DataFrame
        Returns a Spark DataFrame
    """
    train_json = os.path.join(
        dataset_root, "annotations", "instances_train2017.json"
    )
    val_json = os.path.join(
        dataset_root, "annotations", "instances_val2017.json"
    )

    categories = load_categories(train_json)

    examples = []
    for split, anno_file in zip(["train", "val"], [train_json, val_json]):
        coco = COCO(annotation_file=anno_file)
        # Coco has native dependencies, so we do not distributed them
        # to the workers.
        image_ids = coco.imgs
        if limit > 0:
            image_ids = islice(image_ids, limit)
        for image_id in image_ids:
            ann_id = coco.getAnnIds(imgIds=image_id)
            annotations = coco.loadAnns(ann_id)
            annos = []
            for ann in annotations:
                bbox = Box2d.from_top_left(*ann["bbox"])
                annos.append(
                    {
                        "label_id": ann["category_id"],
                        "label": categories[ann["category_id"]]["name"],
                        "bbox": bbox,
                        "area": float(ann["area"]),
                    }
                )
            image_payload = coco.loadImgs(ids=image_id)[0]
            example = {
                "image_id": image_id,
                "annotations": annos,
                "image": Image(
                    os.path.abspath(
                        os.path.join(
                            dataset_root,
                            "{}2017".format(split),
                            image_payload["file_name"],
                        )
                    )
                ),
                "split": split,
            }
            examples.append(example)

    schema = StructType(
        [
            StructField("image_id", LongType(), False),
            StructField(
                "annotations",
                ArrayType(
                    StructType(
                        [
                            StructField("label_id", IntegerType()),
                            StructField("label", StringType()),
                            StructField("area", FloatType()),
                            StructField("bbox", Box2dType()),
                        ]
                    )
                ),
                False,
            ),
            StructField("image", ImageType(), False),
            StructField("split", StringType(), False),
        ]
    )
    df = spark.createDataFrame(examples, schema=schema)
    if asset_dir:
        asset_dir = asset_dir if asset_dir.endswith("/") else asset_dir + "/"
        print("ASSET DIR: ", asset_dir)
        df = df.withColumn("image", image_copy(col("image"), lit(asset_dir)))
    return df



#  Copyright 2021 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

from typing import Tuple

import numpy as np


def encode(arr: np.ndarray) -> np.array:
    """Run-length encoding a matrix.

    Parameters
    ----------
    arr : a data array or n-D metrix/tensor.
    """
    if len(arr.shape) > 1:
        return encode(arr.reshape(-1))

    if len(arr) == 0:
        return []
    total = len(arr)
    conti_idx = np.r_[0, np.flatnonzero(~np.equal(arr[1:], [arr[:-1]])) + 1]
    counts = np.diff(np.r_[conti_idx, total])
    if arr[0]:
        counts = np.insert(counts, 0, 0)
    return counts


def decode(
    rle: np.array, shape: Tuple[int] | Tuple[int, int], order: str = "C"
) -> np.ndarray:
    """Decode (COCO) RLE encoding into a numpy mask.

    Parameters
    ----------
    rle : np.array
        A 1-D array of RLE encoded data.
    shape: tuple of ints
        (height, width)
    order: str
        Numpy array order. If uses Coco-style RLE, order should set to F.
    """
    val = 0
    start_idx = 0
    n = np.sum(rle)
    arr = np.full(n, 0, dtype=np.uint8)
    for length in rle:
        arr[start_idx: start_idx + length] = val  # fmt:skip
        start_idx += length
        val = 1 - val

    return arr.reshape(shape, order=order)



#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Vision Related User-defined Types:

- :py:class:`Image`
"""

from __future__ import annotations

import base64
from io import BytesIO, IOBase
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import List, Optional, Sequence, Union
from urllib.parse import urlparse

# Third-party libraries
import numpy as np
from PIL import Image as PILImage

# Rikai
from rikai.conf import options
from rikai.internal.uri_utils import normalize_uri
from rikai.io import copy, open_output_stream
from rikai.mixin import Asset, Displayable, Drawable, ToDict, ToNumpy, ToPIL
from rikai.spark.types import ImageType
from rikai.types.geometry import Box2d
from rikai.viz import Draw, PILRenderer

__all__ = ["Image"]


class Image(ToNumpy, ToPIL, Asset, Displayable, ToDict):
    """An external Image Asset.

    It contains a reference URI to an image stored on the remote system.

    Parameters
    ----------
    image : bytes, file-like object, str or :py:class:`~pathlib.Path`
        It can be the content of image, or a URI / Path of an image.
    """

    __UDT__ = ImageType()

    def __init__(
        self,
        image: Union[bytes, bytearray, IOBase, str, Path],
    ):
        data, uri = None, None
        if isinstance(image, IOBase):
            data = image.read()
        elif isinstance(image, (bytes, bytearray)):
            data = image
        else:
            uri = image
        super().__init__(data=data, uri=uri)

    @classmethod
    def from_array(
        cls,
        array: np.ndarray,
        uri: Optional[Union[str, Path]] = None,
        mode: Optional[str] = None,
        format: Optional[str] = None,
        **kwargs,
    ) -> Image:
        """Create an image in memory from numpy array.

        Parameters
        ----------
        array : np.ndarray
            Array data
        uri : str or Path
            The external URI to store the data.
        mode : str, optional
            The mode which PIL used to create image. See supported
            `modes on PIL document <https://pillow.readthedocs.io/en/stable/handbook/concepts.html#concept-modes>`_.
        format : str, optional
            The image format to save as. See
            `supported formats <https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.save>`_ for details.
        kwargs : dict, optional
            Optional arguments to pass to `PIL.Image.save <https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.save>`_.

        See Also
        --------
        :py:class:`PIL.Image.fromarray`
        :py:func:`~rikai.spark.functions.vision.numpy_to_image`

        """  # noqa: E501

        assert array is not None
        with PILImage.fromarray(array, mode=mode) as img:
            return cls.from_pil(img, uri, format=format, **kwargs)

    @staticmethod
    def read(
        uri: Union[str, Path],
    ) -> Image:
        """Create an embedded image from external URI

        Parameters
        ----------
        uri : str or Path
            The URI pointed to an image.
        """
        return Image(uri).to_embedded()

    @staticmethod
    def from_pil(
        img: PILImage,
        uri: Optional[Union[str, Path]] = None,
        format: Optional[str] = None,
        **kwargs,
    ) -> Image:
        """Create an image in memory from a :py:class:`PIL.Image`.

        Parameters
        ----------
        img : :py:class:`PIL.Image`
            An PIL Image instance
        uri : str or Path
            The URI to store the image externally.
        format : str, optional
            The image format to save as. See
            `supported formats <https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.save>`_ for details.
        kwargs : dict, optional
            Optional arguments to pass to `PIL.Image.save <https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.save>`_.
        """  # noqa: E501

        format = format if format else options.rikai.image.default.format

        if uri is None:
            buf = BytesIO()
            img.save(buf, format=format, **kwargs)
            return Image(buf.getvalue())

        parsed = urlparse(normalize_uri(uri))
        if parsed.scheme == "file":
            img.save(uri, format=format, **kwargs)
        else:
            with NamedTemporaryFile() as fobj:
                img.save(fobj, format=format, **kwargs)
                fobj.flush()
                copy(fobj.name, uri)
        return Image(uri)

    def display(self, **kwargs):
        """
        Custom visualizer for this image in jupyter notebook

        Parameters
        ----------
        kwargs: dict
            Optional display arguments

        Returns
        -------
        img: IPython.display.Image
        """
        from IPython.display import Image

        if not self.is_embedded and self.uri.startswith("http"):
            return Image(url=self.uri, **kwargs)
        else:
            # Using Data URIs for Databricks Notebook
            with self.open() as fobj:
                data = fobj.read()
                inferred_format = Image(data).format
                encoded = base64.b64encode(data).decode("utf-8")
                url = f"data:image;base64,{encoded}"
                return Image(url=url, format=inferred_format)

    def draw(self, drawable: Union[Drawable, list[Drawable]]) -> Draw:
        return ImageDraw(self).draw(drawable)

    def __repr__(self) -> str:
        if self.is_embedded:
            return "Image(<embedded>)"
        return f"Image(uri={self.uri})"

    def _repr_mimebundle_(self, include=None, exclude=None):
        """default visualizer for embedded mime bundle"""
        return self.display()._repr_mimebundle_(
            include=include, exclude=exclude
        )

    def __eq__(self, other) -> bool:
        return isinstance(other, Image) and super().__eq__(other)

    def __or__(self, other: Drawable) -> Draw:
        """Override ``|`` operator to chain images with
        visualization components.
        """
        return self.draw(other)

    def to_pil(self) -> PILImage:
        """Return an PIL image.

        Note
        ----
        The caller should close the image.
        https://pillow.readthedocs.io/en/stable/reference/open_files.html#image-lifecycle
        """
        return PILImage.open(self.open()).convert("RGB")

    def to_numpy(self) -> np.ndarray:
        """Convert this image into an :py:class:`numpy.ndarray`."""
        with self.to_pil() as pil_img:
            return np.asarray(pil_img)

    def to_dict(self) -> dict:
        if self.is_embedded:
            return {"data": base64.b64encode(self.data).decode("ascii")}
        return {"uri": self.uri}

    def to_embedded(self) -> Image:
        """Convert this image into an embedded image."""
        if self.is_embedded:
            return self
        with self.open() as img:
            return Image(img.read())

    def save(self, uri: Union[str, Path]) -> Image:
        """Save the image into a file, specified by the file path or URI.

        Parameters
        ----------
        uri : str or Path
            The external URI to store the image to.

        Returns
        -------
        :py:class:`Image`
            A new image with the new URI / path
        """
        uri = str(uri)
        if self.is_embedded:
            with open_output_stream(str(uri)) as fobj:
                fobj.write(self.data)
        else:
            copy(self.uri, uri)
        return Image(uri)

    def crop(
        self, box: Union[Box2d, List[Box2d]], format: Optional[str] = None
    ) -> Union[Image, List[Image]]:
        """Crop image specified by the bounding boxes, and returns the cropped
        images.

        Support crop images in batch, to save I/O overhead to download
        the original image.

        Parameters
        ----------
        box : :py:class:`Box2d` or :py:class:`List[Box2d]`
            The bounding box(es) to crop out of this image.
        format : str, optional
            The image format to save as

        Returns
        -------
        :py:class:`Image` or a list of :py:class:`Image`
        """
        if isinstance(box, Box2d):
            with self.to_pil() as pil_image:
                return Image.from_pil(pil_image.crop(box))

        assert isinstance(box, Sequence)
        crops = []
        with self.to_pil() as pil_image:
            for bbox in box:
                with pil_image.crop(bbox) as patch:
                    crops.append(Image.from_pil(patch))
        return crops


class ImageDraw(Draw):
    def __init__(self, img: Image):
        super().__init__()
        self.img = img.to_pil()

    def to_image(self) -> Image:
        if not self.layers:
            raise ValueError("Can not render empty displayable draw")

        render = PILRenderer(self.img)
        for layer in self.layers:
            layer._render(render)
        return Image.from_pil(render.image)

    def display(self, **kwargs) -> "IPython.display.Image":
        """
        Custom visualizer for this ImageDraw in jupyter notebook

        Parameters
        ----------
        kwargs: dict
            Optional display arguments

        Returns
        -------
        img: IPython.display.Image
        """
        return self.to_image().display(**kwargs)



#  Copyright 2021 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Iterator

import numpy as np
import pandas as pd
from pyspark.serializers import CloudPickleSerializer
from pyspark.sql.functions import pandas_udf
from pyspark.sql.types import BinaryType

__all__ = ["generate_udf"]

_pickler = CloudPickleSerializer()


def generate_udf(spec: "rikai.spark.sql.codegen.base.ModelSpec"):
    """Construct a UDF to run sklearn model.

    Parameters
    ----------
    spec : ModelSpec
        the model specifications object

    Returns
    -------
    A Spark Pandas UDF.
    """

    def predict(model, X):
        if hasattr(model, "predict"):
            return model.predict(X)
        elif hasattr(model, "transform"):
            return model.transform(X)
        else:
            raise RuntimeError("predict or transform is not available")

    def sklearn_inference_udf(
        iter: Iterator[pd.Series],
    ) -> Iterator[pd.Series]:
        model = spec.load_model()
        for series in list(iter):
            X = np.vstack(series.apply(_pickler.loads).to_numpy())
            y = [_pickler.dumps(pred.tolist()) for pred in predict(model, X)]
            yield pd.Series(y)

    return pandas_udf(sklearn_inference_udf, returnType=BinaryType())



#  Copyright 2022 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import importlib
import warnings
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, Optional, TypeVar

from jsonschema.exceptions import ValidationError
from jsonschema.validators import validate

from rikai.internal.reflection import find_func
from rikai.logging import logger
from rikai.spark.sql.exceptions import SpecError
from rikai.spark.sql.schema import parse_schema

__all__ = ["ModelSpec", "ModelType", "AnonymousModelType"]


M = TypeVar("M")  # Model Type


# JSON schema specification for the model payload specifications
# used to validate model spec input
SPEC_PAYLOAD_SCHEMA = {
    "type": "object",
    "properties": {
        "version": {
            "type": "string",
            "description": "Model SPEC format version",
        },
        "name": {"type": "string", "description": "Model name"},
        "schema": {"type": "string"},
        "model": {
            "type": "object",
            "description": "model description",
            "properties": {
                "uri": {"type": "string"},
                "flavor": {"type": "string"},
                "model_type": {"type": "string"},
            },
            "required": ["uri"],
        },
        "transforms": {
            "type": "object",
            "properties": {
                "pre": {"type": "string"},
                "post": {"type": "string"},
            },
        },
    },
    "required": ["version", "model"],
}


def _identity(x):
    return x


def is_fully_qualified_name(name: str) -> bool:
    return "." in name


def parse_model_type(flavor: str, model_type: str):
    model_modules_candidates = []
    if is_fully_qualified_name(model_type):
        model_modules_candidates.append(model_type)
    else:
        model_modules_candidates.extend(
            [
                f"rikai.{flavor}.models.{model_type}",
                f"rikai.contrib.{flavor}.models.{model_type}",
            ]
        )
    for model_module in model_modules_candidates:
        try:
            return find_func(f"{model_module}.MODEL_TYPE")
        except ModuleNotFoundError:
            pass
    else:
        raise ModuleNotFoundError(
            f"Model spec not found for model: {model_type}/{flavor}"
        )


class ModelSpec(ABC):
    """Model Spec Payload

    Parameters
    ----------
    spec : dict
        Dictionary representation of an input spec
    validate : bool, default True.
        Validate the spec during construction. Default ``True``.
    """

    def __init__(self, spec: Dict[str, Any], validate: bool = True):
        self._spec = spec
        self._spec["options"] = self._spec.get("options", {})
        if validate:
            self.validate()

    def validate(self):
        """Validate model spec

        Raises
        ------
        SpecError
            If the spec is not well-formatted.
        """
        logger.debug("Validating spec: %s", self._spec)
        try:
            validate(instance=self._spec, schema=SPEC_PAYLOAD_SCHEMA)
        except ValidationError as e:
            raise SpecError(e.message) from e

    @property
    def version(self) -> str:
        """Returns spec version."""
        return str(self._spec["version"])

    @property
    def name(self) -> str:
        """Return model name"""
        return self._spec["name"]

    @property
    def model_uri(self) -> str:
        """Return Model artifact URI"""
        return self._spec["model"]["uri"]

    @property
    def model_type(self) -> "ModelType":
        """Return model type"""
        mtype = self._spec["model"].get("type", None)
        if mtype:
            return parse_model_type(self.flavor, mtype)
        return AnonymousModelType(
            self._spec.get("schema", None),
            self.pre_processing,
            self.post_processing,
        )

    @abstractmethod
    def load_model(self) -> Any:
        """Load the model artifact specified in this spec"""

    @property
    def flavor(self) -> str:
        """Model flavor"""
        return self._spec["model"].get("flavor", "")

    @property
    def schema(self) -> str:
        """Return the output schema of the model."""
        if "schema" in self._spec:
            return parse_schema(self._spec["schema"])
        return self.model_type.schema

    @property
    def options(self) -> Dict[str, Any]:
        """Model options"""
        return self._spec["options"]

    @property
    def pre_processing(self) -> Optional[Callable]:
        """Return pre-processing transform if exists"""
        if (
            "transforms" not in self._spec
            or "pre" not in self._spec["transforms"]
            or self._spec["transforms"]["pre"] is None
        ):
            # Passthrough
            return _identity
        f = find_func(self._spec["transforms"]["pre"])
        return f(self.options)

    @property
    def post_processing(self) -> Optional[Callable]:
        """Return post-processing transform if exists"""
        if (
            "transforms" not in self._spec
            or "post" not in self._spec["transforms"]
            or self._spec["transforms"]["post"] is None
        ):
            # Passthrough
            return _identity
        f = find_func(self._spec["transforms"]["post"])
        return f(self.options)


class ModelType(ABC):
    """Declare a Rikai-compatible Model Type."""

    @abstractmethod
    def load_model(self, spec: ModelSpec, **kwargs):
        """Lazy loading the model from a :class:`ModelSpec`."""
        pass

    @abstractmethod
    def schema(self) -> str:
        """Return the string value of model schema.

        Examples
        --------

        >>> model_type.schema()
        ... "array<struct<box:box2d, score:float, label_id:int>>"
        """
        pass

    def dataType(self) -> "pyspark.sql.types.DataType":
        return parse_schema(self.schema())

    @abstractmethod
    def transform(self) -> Callable:
        """A callable to pre-process the data before calling inference.

        It will be feed into :class:`torch.data.DataLoader` or
        :class:`tensorflow.data.Dataset.map`.

        """
        pass

    @abstractmethod
    def predict(self, *args, **kwargs) -> Any:
        """Run model inference and convert return types into
        Rikai-compatible types.

        """
        pass

    def __call__(self, *args, **kwargs) -> Any:
        return self.predict(*args, **kwargs)

    def release(self):
        """Release underneath resources if applicable.

        It will be called after a model runner finishes a partition in Spark.
        """
        # Noop by default
        pass


class AnonymousModelType(ModelType):
    def __init__(
        self,
        schema: str,
        pre_processing: Optional[Callable],
        post_processing: Optional[Callable],
    ):
        self._schema = schema
        self.pre_processing = pre_processing
        self.post_processing = post_processing
        self.model: Optional[M] = None
        self.spec: Optional[ModelSpec] = None

        warnings.warn(
            "Using schema and pre_processing/post_processing explicitly"
            "is deprecated and will be removed in Rikai 0.2. "
            "Please migrate to a concrete ModelType."
        )

    def schema(self) -> str:
        return self._schema

    def load_model(self, spec: ModelSpec, **kwargs):
        self.model = spec.load_model()
        self.spec = spec

    def transform(self) -> Callable:
        """Adaptor for the pre-processing."""
        return self.pre_processing

    def predict(self, *args, **kwargs) -> Any:
        """Predict combines model inference and post-processing that
        converts inference outputs into Rikai types.

        """
        batch = self.model(*args, **kwargs)
        if self.post_processing:
            batch = self.post_processing(batch)
        return batch



# Generated from src/main/antlr4/org/apache/spark/sql/ml/parser/RikaiModelSchema.g4 by ANTLR 4.8
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .RikaiModelSchemaParser import RikaiModelSchemaParser
else:
    from RikaiModelSchemaParser import RikaiModelSchemaParser

# This class defines a complete generic visitor for a parse tree produced by RikaiModelSchemaParser.

class RikaiModelSchemaVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by RikaiModelSchemaParser#schema.
    def visitSchema(self, ctx:RikaiModelSchemaParser.SchemaContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RikaiModelSchemaParser#unquotedIdentifier.
    def visitUnquotedIdentifier(self, ctx:RikaiModelSchemaParser.UnquotedIdentifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RikaiModelSchemaParser#structType.
    def visitStructType(self, ctx:RikaiModelSchemaParser.StructTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RikaiModelSchemaParser#arrayType.
    def visitArrayType(self, ctx:RikaiModelSchemaParser.ArrayTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RikaiModelSchemaParser#nestedStructType.
    def visitNestedStructType(self, ctx:RikaiModelSchemaParser.NestedStructTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RikaiModelSchemaParser#nestedArrayType.
    def visitNestedArrayType(self, ctx:RikaiModelSchemaParser.NestedArrayTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RikaiModelSchemaParser#plainFieldType.
    def visitPlainFieldType(self, ctx:RikaiModelSchemaParser.PlainFieldTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by RikaiModelSchemaParser#structField.
    def visitStructField(self, ctx:RikaiModelSchemaParser.StructFieldContext):
        return self.visitChildren(ctx)



del RikaiModelSchemaParser


#  Copyright 2021 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Geometry types
"""

from __future__ import annotations

from enum import Enum
from numbers import Real
from typing import Optional, Sequence, Tuple, Union

import numpy as np
from PIL import Image, ImageDraw

from rikai.mixin import Drawable, ToDict, ToNumpy
from rikai.spark.types.geometry import (
    Box2dType,
    Box3dType,
    MaskType,
    PointType,
)
from rikai.types import rle

__all__ = ["Point", "Box3d", "Box2d", "Mask"]


class Point(ToNumpy, ToDict):
    """Point in a 3-D space, specified by ``(x, y, z)`` coordinates.

    Attributes
    ----------
    x : float
        The X coordinate.
    y : float
        The Y coordinate.
    z : float
        The Z coordinate.
    """

    __UDT__ = PointType()

    def __init__(self, x: float, y: float, z: float):
        # pylint: disable=invalid-name
        self.x = float(x)
        self.y = float(y)
        self.z = float(z)

    def __repr__(self) -> str:
        return f"Point({self.x}, {self.y}, {self.z})"

    def __eq__(self, o: object) -> bool:
        return (
            isinstance(o, Point)
            and self.x == o.x
            and self.y == o.y
            and self.z == o.z
        )

    def to_numpy(self) -> np.ndarray:
        return np.array([self.x, self.y, self.z])

    def to_dict(self) -> dict:
        return {"x": self.x, "y": self.y, "z": self.z}


class Box2d(ToNumpy, Sequence, ToDict, Drawable):
    """2-D Bounding Box, defined by ``(xmin, ymin, xmax, ymax)``

    Attributes
    ----------
    xmin : float
        X-coordinate of the top-left point of the box.
    ymin : float
        Y-coordinate of the top-left point of the box.
    xmax : float
        X-coordinate of the bottm-right point of the box.
    ymax : float
        Y-coordinate of the bottm-right point of the box.

    Example
    -------

    >>> box = Box2d(1, 2, 3, 4)
    >>> box / 2
    Box2d(xmin=0.5, ymin=1.0, xmax=1.5, ymax=2.0)
    >>> box * (3.5, 5)
    Box2d(xmin=3.5, ymin=10.0, xmax=10.5, ymax=20.0)
    >>> # Box2d can be used directly with PIL.ImageDraw
    >>> draw = PIL.ImageDraw.Draw(img)
    >>> draw.rectangle(box, fill="green", width=2)
    """

    __UDT__ = Box2dType()

    def __init__(self, xmin: float, ymin: float, xmax: float, ymax: float):
        assert (
            0 <= xmin <= xmax
        ), f"xmin({xmin}) and xmax({xmax}) must satisfy 0 <= xmin <= xmax"
        assert (
            0 <= ymin <= ymax
        ), f"ymin({ymin}) and ymax({ymax}) must satisfy 0 <= ymin <= ymax"
        self.xmin = float(xmin)
        self.ymin = float(ymin)
        self.xmax = float(xmax)
        self.ymax = float(ymax)

    @classmethod
    def from_center(
        cls, center_x: float, center_y: float, width: float, height: float
    ) -> Box2d:
        """Factory method to construct a :py:class:`Box2d` from
        the center point coordinates: ``{center_x, center_y, width, height}``.


        Parameters
        ----------
        center_x : float
            X-coordinate of the center point of the box.
        center_y : float
            Y-coordinate of the center point of the box.
        width : float
            The width of the box.
        height : float
            The height of the box.

        Return
        ------
        Box2d
        """
        assert (
            width >= 0 and height >= 0
        ), f"Box2d width({width}) and height({height}) must be non-negative."
        return Box2d(
            center_x - width / 2,
            center_y - height / 2,
            center_x + width / 2,
            center_y + height / 2,
        )

    @classmethod
    def from_top_left(
        cls, xmin: float, ymin: float, width: float, height: float
    ) -> Box2d:
        """Construct a :py:class:`Box2d` from
        the top-left based coordinates: ``{x0, y0, width, height}``.

        Top-left corner of an image / bbox is `(0, 0)`.

        Several public datasets, including `Coco Dataset`_, use this
        coordinations.

        Parameters
        ----------
        xmin : float
            X-coordinate of the top-left point of the box.
        ymin : float
            Y-coordinate of the top-left point of the box.
        width : float
            The width of the box.
        height : float
            The height of the box.


        References
        ----------
        - `Coco Dataset`_

        .. _Coco Dataset: https://cocodataset.org/
        """
        assert (
            width >= 0 and height >= 0
        ), f"Box2d width({width}) and height({height}) must be non-negative."
        return Box2d(xmin, ymin, xmin + width, ymin + height)

    def __repr__(self) -> str:
        return (
            f"Box2d(xmin={self.xmin}, ymin={self.ymin}, xmax={self.xmax}"
            + f", ymax={self.ymax})"
        )

    def __eq__(self, o: Box2d) -> bool:
        return isinstance(o, Box2d) and np.array_equal(
            self.to_numpy(), o.to_numpy()
        )

    def __len__(self) -> int:
        return 4

    def __getitem__(self, key: int) -> float:
        return [self.xmin, self.ymin, self.xmax, self.ymax][key]

    @staticmethod
    def _verified_scale(
        scale: Union[Real, Tuple[float, float]]
    ) -> Tuple[float, float]:
        if isinstance(scale, Real):
            assert scale > 0, f"scale must be positive, got {scale}"
            return scale, scale
        assert (
            type(scale) == tuple and len(scale) == 2
        ), f"scale must be either a number or a 2-element tuple, got {scale}"
        assert (
            scale[0] > 0 and scale[1] > 0
        ), f"scale must be positive, got {scale}"
        return scale

    def __truediv__(self, scale: Union[int, float, Tuple]) -> Box2d:
        """Scale down :py:class:`Box2d`

        Parameters
        ----------
        scale : number or a 2-element tuple/list
            Scale the Box2d by this amount.

        Return
        ------
        Box2d
            The scaled-down Box2d.

        """
        x_scale, y_scale = self._verified_scale(scale)
        return Box2d(
            xmin=self.xmin / x_scale,
            ymin=self.ymin / y_scale,
            xmax=self.xmax / x_scale,
            ymax=self.ymax / y_scale,
        )

    def __mul__(self, scale: Union[int, float, Tuple]) -> Box2d:
        """Scale up :py:class:`Box2d`

        Parameters
        ----------
        scale : number or a 2-element tuple/list
            Scale the Box2d by this amount.

        Return
        ------
        Box2d
            The scaled-up Box2d.
        """
        x_scale, y_scale = self._verified_scale(scale)
        return self / (1.0 / x_scale, 1.0 / y_scale)

    def _render(self, render: "rikai.viz.Renderer", **kwargs) -> None:
        render.rectangle(self, **kwargs)

    def to_numpy(self) -> np.ndarray:
        """Convert a :py:class:`Box2d` to numpy ndarray:
        ``array([xmin, ymin, xmax, ymax])``

        """
        return np.array([self.xmin, self.ymin, self.xmax, self.ymax])

    def to_dict(self) -> dict:
        return {
            "xmin": self.xmin,
            "ymin": self.ymin,
            "xmax": self.xmax,
            "ymax": self.ymax,
        }

    @property
    def width(self) -> float:
        return self.xmax - self.xmin

    @property
    def height(self) -> float:
        return self.ymax - self.ymin

    @property
    def area(self) -> float:
        """Area of the bounding box"""
        return self.width * self.height

    @staticmethod
    def _area(arr: np.ndarray) -> np.array:
        """Calculate box area from an array of boxes."""
        return np.maximum(0, arr[:, 2] - arr[:, 0]) * np.maximum(
            0, arr[:, 3] - arr[:, 1]
        )

    @staticmethod
    def ious(
        boxes1: Union[Sequence[Box2d], np.ndarray],
        boxes2: Union[Sequence[Box2d], np.ndarray],
    ) -> np.ndarray:
        """Compute intersection over union(IOU).

        Parameters
        ----------
        boxes1 : :py:class:`numpy.ndarray`
            a list of Box2d with length of N
        boxes2 : :py:class:`numpy.ndarray`
            a list of Box2d with length of M
        Return
        ------
        :py:class:`numpy.ndarray`
             For two lists of box2ds, which have the length of N, and M respectively, this function should return a N*M matrix,
              each element is the iou value (float,[0, 1]).

        Example
        -------

        >>> import random
        >>>
        >>> def a_random_box2d():
        ...   x_min = random.uniform(0, 1)
        ...   y_min = random.uniform(0, 1)
        ...   x_max = random.uniform(x_min, 1)
        ...   y_max = random.uniform(y_min, 1)
        ...  return Box2d(x_min, y_min, x_max, y_max)
        >>>
        >>> list1 = [a_random_box2d() for _ in range(0, 2)]
        >>>
        >>> list2 = [a_random_box2d() for _ in range(0, 3)]
        >>>
        >>> Box2d.ious(list1, list2)
        """  # noqa: E501

        assert isinstance(boxes1, (Sequence, np.ndarray))
        assert isinstance(boxes2, (Sequence, np.ndarray))

        if not isinstance(boxes1, np.ndarray):
            boxes1 = np.array(boxes1)
        if not isinstance(boxes2, np.ndarray):
            boxes2 = np.array(boxes2)
        row_count = boxes1.shape[0]
        area1 = Box2d._area(boxes1).reshape(row_count, -1)
        area2 = Box2d._area(boxes2)

        xmin = np.maximum(boxes1[:, 0].reshape((row_count, -1)), boxes2[:, 0])
        ymin = np.maximum(boxes1[:, 1].reshape((row_count, -1)), boxes2[:, 1])
        xmax = np.minimum(boxes1[:, 2].reshape((row_count, -1)), boxes2[:, 2])
        ymax = np.minimum(boxes1[:, 3].reshape((row_count, -1)), boxes2[:, 3])

        inter_area = np.maximum(0, xmax - xmin) * np.maximum(0, ymax - ymin)

        iou_mat = inter_area / (area1 + area2 - inter_area)

        return iou_mat

    def iou(
        self, other: Union[Box2d, Sequence[Box2d], np.ndarray]
    ) -> Union[float, np.ndarray]:
        """Compute intersection over union(IOU)."""
        assert isinstance(
            other, (Box2d, Sequence, np.ndarray)
        ), f"Can only compute iou between Box2d, got {type(other)}"
        if isinstance(other, Box2d):
            other_arr = np.array([other])
        else:
            other_arr = np.array(other)
        if other_arr.size == 0:
            return np.zeros(0)

        self_arr = np.array(self)
        # Find intersection
        xmin = np.maximum(self_arr[0], other_arr[:, 0])
        ymin = np.maximum(self_arr[1], other_arr[:, 1])
        xmax = np.minimum(self_arr[2], other_arr[:, 2])
        ymax = np.minimum(self_arr[3], other_arr[:, 3])
        inter_arr = np.array([xmin, ymin, xmax, ymax]).T
        inter_area = self._area(inter_arr)

        iou_arr = inter_area / (
            self._area(np.array([self_arr]))
            + self._area(other_arr)
            - inter_area
        )
        if isinstance(other, Box2d):
            return iou_arr[0]
        return iou_arr


class Box3d(ToNumpy, ToDict):
    """A 3-D bounding box

    Attributes
    ----------
    center : Point
        Center :py:class:`Point` of the bounding box
    length : float
        The x dimention of the box
    width : float
        The y dimention of the box
    height : float
        The z dimention of the box
    heading : float
        The heading of the bounding box (in radians).  The heading is the angle
        required to rotate +x to the surface normal of the box front face. It is
        normalized to ``[-pi, pi)``.

    References
    ----------
    * Waymo Dataset Spec https://github.com/waymo-research/waymo-open-dataset/blob/master/waymo_open_dataset/label.proto
    """  # noqa: E501

    __UDT__ = Box3dType()

    def __init__(
        self,
        center: Point,
        length: float,
        width: float,
        height: float,
        heading: float,
    ):
        self.center = center
        self.length = float(length)
        self.width = float(width)
        self.height = float(height)
        self.heading = float(heading)

    def __repr__(self) -> str:
        return (
            f"Box3d(center={self.center}, l={self.length}, "
            + f"h={self.height}, w={self.width}, heading={self.heading})"
        )

    def __eq__(self, o: object) -> bool:
        return (
            isinstance(o, Box3d)
            and self.center == o.center
            and self.length == o.length
            and self.width == o.width
            and self.height == o.height
            and self.heading == o.heading
        )

    def to_numpy(self) -> np.ndarray:
        return None

    def to_dict(self) -> dict:
        return {
            "center": self.center,
            "length": self.length,
            "width": self.width,
            "height": self.height,
            "heading": self.heading,
        }


class Mask(ToNumpy, ToDict, Drawable):
    """2-d Mask over an image

    This 2D mask can be built from:

    - A binary-valued (0 or 1) 2D-numpy matrix.
    - A Run Length Encoded (RLE) data. It supports both row-based RLE
      (:py:class:`Mask.Type.RLE`) or column-based RLE
      (:py:class:`Mask.Type.COCO_RLE`) which is used in the Coco dataset.
    - A Polygon ``[x0, y0, x1, y1, ..., xn, yn]``

    Parameters
    ----------
    data: list or :py:class:`np.ndarray`
        The mask data. Can be a numpy array or a list.
    height: int, optional
        The height of the image this mask applies to.
    width: int, optional
        The width of the image this mask applies to.
    mask_type: :py:class:`Mask.Type`
        The type of the mask.

    Examples
    --------

    .. code-block:: python

        from pycocotools.coco import COCO
        from rikai.types import Mask

        coco = COCO("instance_train2017.json")
        ann = coco.loadAnns(ann_id)
        image = coco.loadImgs(ann["image_id"])
        if ann["iscrowed"] == 0:
            mask = Mask.from_polygon(
                ann["segmentation"],
                height=image["height"],
                width=image["width],
            )
        else:
            mask = Mask.from_coco_rle(
                ann["segmentation"]["counts"],
                height=image["height"],
                width=image["width],
            )

    """

    __UDT__ = MaskType()

    class Type(Enum):
        """Mask type."""

        POLYGON = 1
        RLE = 2
        COCO_RLE = 3  # COCO style RLE, column-based

    def __init__(
        self,
        data: Union[list, np.ndarray],
        height: Optional[int] = None,
        width: Optional[int] = None,
        mask_type: Mask.Type = Type.POLYGON,
    ):
        if mask_type != Mask.Type.POLYGON and (
            height is None or width is None
        ):
            raise ValueError("Must provide height and width for RLE type")

        self.type = mask_type
        self.data = data

        self.height = height
        self.width = width

    @staticmethod
    def from_rle(data: list[int], height: int, width: int) -> Mask:
        """Convert a (row-based) RLE mask (segmentation) into Mask

        Parameters
        ----------
        data: list[int]
            the RLE data
        height: int
            The height of the image which the mask applies to.
        width: int
            The width of the image which the mask applies to.
        """

        return Mask(data, height=height, width=width, mask_type=Mask.Type.RLE)

    @staticmethod
    def from_coco_rle(data: list[int], height: int, width: int) -> Mask:
        """Convert a COCO RLE mask (segmentation) into Mask

        Parameters
        ----------
        data: list[int]
            the RLE data
        height: int
            The height of the image which the mask applies to.
        width: int
            The width of the image which the mask applies to.
        """
        return Mask(
            data, height=height, width=width, mask_type=Mask.Type.COCO_RLE
        )

    @staticmethod
    def from_polygon(data: list[list[float]], height: int, width: int) -> Mask:
        """Build mask from a Polygon

        Parameters
        ----------
        data: list[list[float]]
            Multiple Polygon segmentation data. i.e.,
            ``[[x0, y0, x1, y1, ...], [x0, y0, x1, y1, ...]])``
        height: int
            The height of the image which the mask applies to.
        width: int
            The width of the image which the mask applies to.
        """
        return Mask(
            data, height=height, width=width, mask_type=Mask.Type.POLYGON
        )

    @staticmethod
    def from_mask(mask: np.ndarray) -> Mask:
        """Build mask from a numpy array.

        Parameters
        ----------
        mask : np.ndarray
            A binary-valued (0/1) numpy array
        """
        assert len(mask.shape) > 1, "Must have more than 2-dimensions"
        return Mask(data=rle.encode(mask), mask_type=Mask.Type.RLE)

    def __repr__(self):
        return f"Mask(type={self.type}, data=...)"

    def __eq__(self, other):
        return (
            isinstance(other, Mask)
            and self.type == other.type
            and self.height == other.height
            and self.width == other.width
            and np.array_equal(self.data, other.data)
        )

    def _polygon_to_mask(self) -> np.ndarray:
        arr = np.zeros((self.height, self.width), dtype=np.uint8)
        with Image.fromarray(arr) as im:
            draw = ImageDraw.Draw(im)
            for polygon in self.data:
                draw.polygon(list(np.array(polygon)), fill=1)
            return np.array(im)

    def _render(self, render, **kwargs):
        """Render a Mask"""
        if self.type == Mask.Type.POLYGON:
            for segmentation in self.data:
                render.polygon(segmentation, **kwargs)
        else:
            render.mask(self.to_mask())

    def to_mask(self) -> np.ndarray:
        """Convert this mask to a numpy array."""
        if self.type == Mask.Type.POLYGON:
            return self._polygon_to_mask()
        elif self.type == Mask.Type.RLE:
            return rle.decode(self.data, shape=(self.height, self.width))
        elif self.type == Mask.Type.COCO_RLE:
            return rle.decode(
                self.data, shape=(self.height, self.width), order="F"
            )
        else:
            raise ValueError("Unrecognized type")

    def to_numpy(self) -> np.ndarray:
        return self.to_mask()

    def to_dict(self) -> dict:
        ret = {
            "type": self.type.value,
            "width": self.width,
            "height": self.height,
            "data": self.data,
        }
        return ret

    def iou(self, other: Mask) -> float:
        this_mask = self.to_mask()
        other_mask = other.to_mask()
        intersection = np.count_nonzero(np.logical_and(this_mask, other_mask))
        union = np.count_nonzero(np.logical_or(this_mask, other_mask))
        try:
            return intersection / union
        except ZeroDivisionError:
            return 0



#  Copyright 2022 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any, Callable, Dict

from torchvision import transforms as T

from rikai.types import Box2d

__all__ = ["pre_processing", "post_processing"]

DEFAULT_MIN_SCORE = 0.5


def pre_processing(options: Dict[str, Any]) -> Callable:
    return T.ToTensor()


def post_processing(options: Dict[str, Any]) -> Callable:
    min_score = float(options.get("min_score", DEFAULT_MIN_SCORE))

    def post_process_func(batch):
        results = []
        for predicts in batch:
            predict_result = []
            for box, label, score in zip(
                predicts["boxes"].tolist(),
                predicts["labels"].tolist(),
                predicts["scores"].tolist(),
            ):
                if score < min_score:
                    continue
                predict_result.append(
                    {
                        "box": Box2d(*box),
                        "label_id": label,
                        "score": score,
                    }
                )

            results.append(predict_result)
        return results

    return post_process_func



#  Copyright (c) 2021 Rikai Authors
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
from typing import Any, Dict, Optional
from urllib.parse import urlparse

try:
    import mlflow
except ImportError:
    raise ImportError(
        "Couldn't import mlflow. Please make sure to "
        "`pip install mlflow` explicitly or install "
        "the correct extras like `pip install rikai[mlflow]`"
    )
from mlflow.tracking import MlflowClient

from rikai.logging import logger
from rikai.spark.sql.codegen.base import ModelSpec, Registry, udf_from_spec
from rikai.spark.sql.codegen.mlflow_logger import (
    CONF_MLFLOW_MODEL_FLAVOR,
    CONF_MLFLOW_MODEL_TYPE,
    CONF_MLFLOW_OUTPUT_SCHEMA,
    CONF_MLFLOW_POST_PROCESSING,
    CONF_MLFLOW_PRE_PROCESSING,
    CONF_MLFLOW_SPEC_VERSION,
    CONF_MLFLOW_TRACKING_URI,
    MlflowLogger,
)
from rikai.spark.sql.exceptions import SpecError

__all__ = ["MlflowRegistry"]


class MlflowModelSpec(ModelSpec):
    """Model Spec.

    Parameters
    ----------
    model_uri: str
        The uri that Mlflow registry knows how to read
    model_conf: dict
        Configurations required to specify the model
    tracking_uri: str
        The mlflow tracking uri
    options: Dict[str, Any], optional
        Additionally model runtime options
    validate: bool, default True.
        Validate the spec during construction. Default ``True``.
    """

    def __init__(
        self,
        model_uri: str,
        model_conf: dict,
        tracking_uri: str,
        options: Optional[Dict[str, Any]] = None,
        validate: bool = True,
    ):
        self.tracking_uri = tracking_uri
        spec_dict = self._load_spec_dict(model_uri, model_conf, options or {})
        super().__init__(spec_dict, validate=validate)
        self._artifact = None

    def load_model(self) -> Any:
        """Load the model artifact specified in this spec"""
        # Currently mlflow model registry load_model is only accessible
        # via fluent API `mlflow.<flavor>.load_model`. So let's be a good
        # samaritan and put the tracking uri back where we found it
        # (just in case)
        old_uri = mlflow.get_tracking_uri()
        try:
            mlflow.set_tracking_uri(self.tracking_uri)
            if self.flavor == "tensorflow":
                return mlflow.pyfunc.load_model(
                    self.model_uri
                )._model_impl.model
            else:
                return getattr(mlflow, self.flavor).load_model(self.model_uri)
        finally:
            mlflow.set_tracking_uri(old_uri)

    def _load_spec_dict(self, uri: str, conf: dict, options: dict) -> dict:
        """Convert the Run into a ModelSpec

        Parameters
        ----------
        uri: str
            Rikai model reference
        conf: dict
            Configurations that specifies the model
        options: dict, default None
            Runtime options to be used by the model/transforms

        Returns
        -------
        spec: Dict[str, Any]
        """
        spec = {
            "version": conf.get(
                CONF_MLFLOW_SPEC_VERSION,
                MlflowLogger._CURRENT_MODEL_SPEC_VERSION,
            ),
            "schema": conf.get(CONF_MLFLOW_OUTPUT_SCHEMA, None),
            "model": {
                "flavor": _get_model_prop(conf, CONF_MLFLOW_MODEL_FLAVOR),
                "uri": uri,
                "type": conf.get(CONF_MLFLOW_MODEL_TYPE, None),
            },
            "transforms": {
                "pre": conf.get(CONF_MLFLOW_PRE_PROCESSING, None),
                "post": conf.get(CONF_MLFLOW_POST_PROCESSING, None),
            },
        }

        # remove none value
        if not spec["transforms"]["pre"]:
            del spec["transforms"]["pre"]
        if not spec["transforms"]["post"]:
            del spec["transforms"]["post"]
        if not spec["model"]["type"]:
            del spec["model"]["type"]
        if not spec["schema"]:
            del spec["schema"]

        # options
        for key, value in conf.items():
            key = key.lower().strip()
            if key.startswith("rikai.option."):
                sub_len = len("rikai.option.")
                options[key[sub_len:]] = value
        if options:
            options.update(options)
        if len(options) > 0:
            spec["options"] = options

        return spec


def _get_model_prop(
    run_tags: dict,
    conf_name: str,
    extra_options: dict = None,
    option_key: str = None,
    default_value: Any = None,
    raise_if_absent: bool = True,
) -> Any:
    extra_options = extra_options or {}
    option_key = option_key or conf_name
    value = run_tags.get(
        conf_name, extra_options.get(option_key, default_value)
    )
    if not value and raise_if_absent:
        raise ValueError(
            (
                "Please use rikai.mlflow.<flavor>.log_model after "
                "training, or specify {} in CREATE MODEL OPTIONS. "
                "Tags: {}. "
                "Options: {}"
            ).format(conf_name, run_tags, extra_options)
        )
    return value


class MlflowRegistry(Registry):
    """MLFlow-based Model Registry"""

    def __init__(self):
        self._mlflow_client = None

    def __repr__(self):
        return "MlflowRegistry"

    @property
    def tracking_client(self):
        if (
            not self._mlflow_client
            or self._mlflow_client._tracking_client.tracking_uri
            != self.mlflow_tracking_uri
        ):
            self._mlflow_client = MlflowClient(self.mlflow_tracking_uri)
        return self._mlflow_client

    @property
    def mlflow_tracking_uri(self):
        return os.environ.get(CONF_MLFLOW_TRACKING_URI)

    def make_model_spec(self, raw_spec):
        uri = raw_spec["uri"]
        parsed = urlparse(uri)
        if parsed.netloc:
            raise ValueError(
                "URI with 2 forward slashes is not supported, "
                "try URI with 1 slash instead"
            )
        if parsed.scheme != "mlflow":
            raise ValueError("Expect schema: mlflow, but got {parsed.scheme}")
        parts = parsed.path.strip("/").split("/", 1)
        model_uri, run = self.get_model_version(*parts)
        spec = MlflowModelSpec(
            model_uri,
            self.get_model_conf(raw_spec, run),
            self.mlflow_tracking_uri,
            options=self.get_options(raw_spec, run),
        )
        return spec

    def get_model_conf(self, spec, run):
        """
        Get the configurations needed to specify the model
        """
        from_spec = [
            (CONF_MLFLOW_MODEL_FLAVOR, spec["flavor"]),
            (CONF_MLFLOW_PRE_PROCESSING, spec.get("preprocessor", None)),
            (CONF_MLFLOW_POST_PROCESSING, spec.get("postprocessor", None)),
            (CONF_MLFLOW_OUTPUT_SCHEMA, spec["schema"]),
        ]
        tags = {k: v for k, v in from_spec if v}
        # PEP 448 syntax, right-to-left priority order
        return {**run.data.tags, **tags}

    def get_options(self, spec, run):
        options = run.data.params
        if spec.get("options", {}):
            options.update(spec.get("options"))
        return options

    def get_model_version(
        self, model, stage_or_version=None
    ) -> (str, mlflow.entities.Run):
        """
        Get the model uri that mlflow model registry understands for loading
        a model and the corresponding Run with metadata needed for the spec
        """
        # TODO allow default stage from config
        stage_or_version = stage_or_version or "none"

        if stage_or_version.isdigit():
            # Pegged to version number
            run_id = self.tracking_client.get_model_version(
                model, int(stage_or_version)
            ).run_id
            version = int(stage_or_version)
        else:
            # Latest version in stage
            results = self.tracking_client.get_latest_versions(
                model, stages=[stage_or_version.lower()]
            )
            if not results:
                msg = "No versions found for model {} in stage {}".format(
                    model, stage_or_version
                )
                raise SpecError(msg)
            run_id, version = results[0].run_id, results[0].version

        run = self.tracking_client.get_run(run_id)
        return "models:/{}/{}".format(model, version), run



#  Copyright 2021 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any, Callable

import torch
from torchvision import transforms as T

from rikai.pytorch.models import TorchModelType

"""
Adapted from https://github.com/pytorch/pytorch.github.io/blob/site/assets/hub/pytorch_vision_resnet.ipynb
"""  # noqa E501


class ResnetModelType(TorchModelType):
    def __repr__(self):
        return "ResnetModelType"

    def schema(self) -> str:
        return "array<float>"

    def transform(self) -> Callable:
        """
        All pre-trained models expect input images normalized in the same way,
        i.e. mini-batches of 3-channel RGB images of shape (3 x H x W), where
        H and W are expected to be at least 224. The images have to be loaded
        in to a range of [0, 1] and then normalized using
        mean = [0.485, 0.456, 0.406] and std = [0.229, 0.224, 0.225].
        """
        return T.Compose(
            [
                T.Resize(256),
                T.CenterCrop(224),
                T.ToTensor(),
                T.Normalize(
                    mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]
                ),
            ]
        )

    def predict(self, images, *args, **kwargs) -> Any:
        assert (
            self.model is not None
        ), "model has not been initialized via load_model"
        batch = self.model(images)
        results = []
        for result in batch:
            results.append(result.detach().cpu().tolist())
        return results


MODEL_TYPE = ResnetModelType()



#  Copyright 2021 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Optional, Tuple, Union

import numpy as np
from PIL import Image as PILImage
from PIL import ImageDraw

from rikai.conf import CONF_RIKAI_VIZ_COLOR, get_option
from rikai.mixin import Displayable, Drawable

__all__ = ["Style", "Text"]


class Style(Drawable):
    """Styling a drawable-component.

    Examples
    --------

    >>> from rikai.viz import Style
    >>> from rikai.types import Box2d, Image
    ...
    >>> img = Image(uri="s3://....")
    >>> bbox1, bbox2 = Box2d(1, 2, 3, 4), Box2d(3, 4, 5, 6)
    >>> bbox_style = Style(color="yellow", width=4)
    >>> image | bbox_style(bbox1) | bbox_style(bbox2)
    """

    def __init__(self, **kwarg):
        self.kwargs = kwarg
        self.inner = None  # type: Optional[list[Drawable]]

    def __repr__(self):
        return f"style({self.kwargs})"

    def __call__(self, inner: Union[Drawable, list[Drawable]]) -> Drawable:
        # Make a copy of Style so the same style can be applied
        # to multiple drawables
        s = Style(**self.kwargs)
        if isinstance(inner, Drawable):
            inner = [inner]
        s.inner = inner
        return s

    def _render(self, render: Renderer, **kwargs):
        if self.inner is None:
            raise ValueError(
                "This style object has not attack to a Drawable yet"
            )
        # TODO: catch excessive parameters
        kwargs.update(self.kwargs)
        for inner_draw in self.inner:
            inner_draw._render(render, **kwargs)


class Draw(Displayable, ABC):
    """Draw is a container that contain the elements for visualized lazily."""

    def __init__(self):
        self.layers = []

    def __repr__(self):
        first_layer = self.layers[0] if self.layers else "N/A"
        return f"Draw({first_layer})"

    def _repr_mimebundle_(self, include=None, exclude=None):
        """default visualizer for embedded mime bundle"""
        return self.display()._repr_mimebundle_(
            include=include, exclude=exclude
        )

    def draw(self, layer: Union[Drawable, list[Drawable]]) -> Draw:
        # layer can not be checked against typing.Sequence or typing.Iterable,
        # because many of the Drawables are iterables (i.e., Box2d).
        if isinstance(layer, Drawable):
            layer = [layer]
        elif not isinstance(layer, (Drawable, list)):
            raise ValueError(
                f"{layer} must be one Drawable or a list of Drawable"
            )
        self.layers.extend(layer)
        return self

    def __or__(self, other: Union[Drawable, list[Drawable]]) -> Draw:
        return self.draw(other)


class Renderer(ABC):
    """The base class for rendering a :py:class:`Draw`."""

    @abstractmethod
    def rectangle(
        self, xy, color: str = get_option(CONF_RIKAI_VIZ_COLOR), width: int = 1
    ):
        pass

    @abstractmethod
    def polygon(self, xy):
        pass

    @abstractmethod
    def text(self, xy, text: str, color: str = ""):
        pass

    @abstractmethod
    def mask(self, arr: np.ndarray):
        pass


class PILRenderer(Renderer):
    """Use PIL to render drawables"""

    def __init__(self, img: PILImage):
        self.img = img.convert("RGBA")
        self.draw = ImageDraw.Draw(self.img)  # type: ImageDraw

    @property
    def image(self) -> PILImage:
        return self.img

    def rectangle(
        self, xy, color: str = get_option(CONF_RIKAI_VIZ_COLOR), width: int = 1
    ):
        self.draw.rectangle(xy, outline=color, width=width)

    def polygon(
        self,
        xy,
        color: str = get_option(CONF_RIKAI_VIZ_COLOR),
        fill: bool = True,
    ):
        if fill:
            overlay = PILImage.new("RGBA", self.img.size, (255, 255, 255, 0))
            overlay_draw = ImageDraw.Draw(overlay)
            mask_img = PILImage.new("L", self.draw.im.size, 0)
            mask_draw = ImageDraw.Draw(mask_img)
            mask_draw.polygon(xy=xy, fill=color)
            overlay_draw.bitmap((0, 0), mask_img, fill=color)

            self.img = PILImage.alpha_composite(self.img, overlay)
            self.draw = ImageDraw.Draw(self.img)
        else:
            self.draw.polygon(xy=xy, outline=color)

    def text(
        self, xy, text: str, color: str = get_option(CONF_RIKAI_VIZ_COLOR)
    ):
        self.draw.text(xy, text, fill=color)

    def mask(
        self, arr: np.ndarray, color: str = get_option(CONF_RIKAI_VIZ_COLOR)
    ):
        overlay = PILImage.new("RGBA", self.img.size, (255, 255, 255, 0))
        overlay_draw = ImageDraw.Draw(overlay)
        overlay_draw.bitmap((0, 0), PILImage.fromarray(arr), fill=color)

        self.img = PILImage.alpha_composite(self.img, overlay)
        self.draw = ImageDraw.Draw(self.img)


class Text(Drawable):
    """Render a Text

    Parameters
    ----------
    text : str
        The text content to be rendered
    xy : Tuple[int, int]
        The location to render the text
    color : str, optional
        The RGB color string to render the text
    """

    def __init__(
        self,
        text: str,
        xy: Tuple[int, int],
        color: str = get_option(CONF_RIKAI_VIZ_COLOR),
    ):
        self.text = text
        self.xy = xy
        self.color = color

    def _render(self, render: Renderer, **kwargs):
        kwargs["color"] = kwargs.get("color", self.color)
        return render.text(self.xy, self.text, **kwargs)



#  Copyright 2021 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Domain-specific Pyspark UDFs
"""

from pyspark.sql import SparkSession

from rikai.spark.functions.geometry import *
from rikai.spark.functions.io import *
from rikai.spark.functions.vision import *

__all__ = ["init"]


def init(spark: SparkSession):
    """Register all rikai UDFs"""
    from rikai.spark.functions import geometry, io, vision

    all_geo_udfs = [
        "area",
        "box2d",
        "box2d_from_center",
        "box2d_from_top_left",
    ]
    all_io_udfs = ["copy"]
    all_vis_udfs = [
        "crop",
        "to_image",
        "image_copy",
        "numpy_to_image",
        "video_to_images",
        "spectrogram_image",
        "video_metadata",
    ]
    for name in all_geo_udfs:
        spark.udf.register(name, getattr(geometry, name))
    for name in all_io_udfs:
        spark.udf.register(name, getattr(io, name))
    for name in all_vis_udfs:
        spark.udf.register(name, getattr(vision, name))



#  Copyright 2022 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Iterator

import pandas as pd
import tensorflow as tf
from pyspark.serializers import CloudPickleSerializer
from pyspark.sql.functions import pandas_udf
from pyspark.sql.types import BinaryType

from rikai.pytorch.pandas import PandasDataset
from rikai.spark.sql.model import ModelSpec
from rikai.types import Image

DEFAULT_BATCH_SIZE = 4


__all__ = ["generate_udf", "load_model_from_uri"]

_pickler = CloudPickleSerializer()


def infer_output_signature(blob, is_udf: bool):
    if is_udf:
        row = _pickler.loads(blob)
    else:
        row = blob

    if isinstance(row, Image):
        image_arr = row.to_numpy()
        if len(image_arr.shape) == 2:
            return tf.TensorSpec(shape=(None, None), dtype=tf.uint8)
        elif len(image_arr.shape) == 3:
            return tf.TensorSpec(
                shape=(
                    None,
                    None,
                    image_arr.shape[2],
                ),
                dtype=tf.uint8,
            )
        else:
            raise ValueError(
                f"Image should only have 2d or 3d shape, got {image_arr.shape}"
            )
    else:
        return tf.TensorSpec.from_tensor(row)


def _generate(payload: ModelSpec, is_udf: bool = True):
    """Construct a UDF to run Tensorflow (Karas) Model"""

    model = payload.model_type
    options = payload.options
    batch_size = int(options.get("batch_size", DEFAULT_BATCH_SIZE))

    def tf_inference_udf(
        iter: Iterator[pd.DataFrame],
    ) -> Iterator[pd.Series]:
        model.load_model(payload)

        signature = None
        for df in iter:
            if signature is None:
                signature = infer_output_signature(df.iloc[0], is_udf)

            ds = PandasDataset(df, unpickle=is_udf, use_pil=True)
            data = tf.data.Dataset.from_generator(
                ds,
                output_signature=tf.TensorSpec(
                    shape=(None, None, 3), dtype=tf.uint8, name="input_tensor"
                ),
            )

            if model.transform():
                data = data.map(model.transform())
            data = data.batch(batch_size)

            results = []
            for batch in data:
                predictions = model(batch)
                results.extend(
                    [_pickler.dumps(p) if is_udf else p for p in predictions]
                )
            yield pd.Series(results)

    if is_udf:
        return pandas_udf(tf_inference_udf, returnType=BinaryType())
    else:
        return tf_inference_udf


def generate_inference_func(payload: ModelSpec):
    return _generate(payload, False)


def generate_udf(payload: ModelSpec):
    return _generate(payload, True)


def load_model_from_uri(uri: str):
    from tensorflow import keras

    return keras.models.load_model(uri)



#  Copyright 2021 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Vision related Spark UDFs.
"""

# Standard library
import os
from pathlib import Path
from typing import List, Optional, Union

# Third Party
import numpy as np
from pyspark.sql.functions import udf
from pyspark.sql.types import (
    ArrayType,
    FloatType,
    IntegerType,
    StringType,
    StructField,
    StructType,
)

# Rikai
from rikai.io import copy as _copy
from rikai.logging import logger
from rikai.numpy import ndarray
from rikai.spark.types.vision import ImageType
from rikai.types.geometry import Box2d
from rikai.types.video import (
    Segment,
    SingleFrameSampler,
    VideoStream,
    YouTubeVideo,
)
from rikai.types.vision import Image

__all__ = [
    "crop",
    "to_image",
    "image_copy",
    "numpy_to_image",
    "video_to_images",
    "spectrogram_image",
    "video_metadata",
    "video_metadata_schema",
]


@udf(returnType=ImageType())
def to_image(image_data: Union[bytes, bytearray, str, Path]) -> Image:
    """Build an :py:class:`Image` from
    bytes, file-like object, str, or :py:class:`~pathlib.Path`.

    Parameters
    ----------
    image_data : bytes, bytearray, str, Path
        The resource identifier or bytes of the source image.

    Return
    ------
    img: Image
        An Image from the given embedded data or URI

    Example
    -------

    >>> df = spark.read.format("image").load("<path-to-data>")
    >>>
    >>> df.withColumn("new_image", to_image("image.data"))
    """
    return Image(image_data)


@udf(returnType=ImageType())
def image_copy(img: Image, uri: str) -> Image:
    """Copy the image to a new destination, specified by the URI.

    Parameters
    ----------
    img : Image
        An image object
    uri : str
        The base directory to copy the image to.

    Return
    ------
    Image
        Return a new image pointed to the new URI
    """
    logger.info("Copying image src=%s dest=%s", img.uri, uri)
    return Image(_copy(img.uri, uri))


@udf(returnType=ImageType())
def numpy_to_image(
    array: ndarray, uri: str, format: str = None, **kwargs
) -> Image:
    """Convert a numpy array to image, and upload to external storage.

    Parameters
    ----------
    array : :py:class:`numpy.ndarray`
        Image data.
    uri : str
        The base directory to copy the image to.
    format : str, optional
        The image format to save as. See
        `supported formats <https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.save>`_ for details.
    kwargs : dict, optional
        Optional arguments to pass to `PIL.Image.save <https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.save>`_.

    Return
    ------
    Image
        Return a new image pointed to the new URI.

    Example
    -------

    >>> spark.createDataFrame(..).registerTempTable("df")
    >>>
    >>> spark.sql(\"\"\"SELECT numpy_to_image(
    ...        resize(grayscale(image)),
    ...        lit('s3://asset')
    ...    ) AS new_image FROM df\"\"\")

    See Also
    --------
    :py:meth:`rikai.types.vision.Image.from_array`
    """  # noqa: E501
    return Image.from_array(array, uri, format=format, **kwargs)


@udf(returnType=ArrayType(ImageType()))
def video_to_images(
    video: Union[VideoStream, YouTubeVideo],
    output_uri: str,
    segment: Segment = Segment(0, -1),
    sample_rate: int = 1,
    max_samples: int = 15000,
    quality: str = "worst",
    image_format: str = "png",
    **image_kwargs,
) -> list:
    """Extract video frames into a list of images.

    Parameters
    ----------
    video : Video
        An video object, either YouTubeVideo or VideoStream.
    output_uri: str
        Frames will be written as <output_uri>/<fno>.<img_format>
    segment: Segment, default Segment(0, -1)
        A Segment object, localizing video in time to (start_fno, end_fno)
    sample_rate : int, default 1
        Keep 1 out of every sample_rate frames.
    max_samples : int, default 15000
        Return at most this many frames (-1 means no max)
    quality: str, default 'worst'
        Either 'worst' (lowest bitrate) or 'best' (highest bitrate)
        See: https://pythonhosted.org/Pafy/index.html#Pafy.Pafy.getbest
    image_format : str, optional
        The image format to save as. See
        `supported formats <https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.save>`_ for details.
    image_kwargs : dict, optional
        Optional arguments to pass to `PIL.Image.save <https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.save>`_.
    ------
    List
        Return a list of images from video indexed by frame number.
    """  # noqa: E501
    assert isinstance(
        video, (YouTubeVideo, VideoStream)
    ), "Input type must be YouTubeVideo or VideoStream"
    assert isinstance(segment, Segment), "Second input type must be Segment"

    start_frame = segment.start_fno
    if segment.end_fno > 0:
        max_samples = min((segment.end_fno - start_frame), max_samples)

    if isinstance(video, YouTubeVideo):
        video_iterator = SingleFrameSampler(
            video.get_stream(quality=quality),
            sample_rate,
            start_frame,
            max_samples,
        )
    else:
        video_iterator = SingleFrameSampler(
            video, sample_rate, start_frame, max_samples
        )

    return [
        Image.from_array(
            img,
            os.path.join(
                output_uri,
                "{}.{}".format(
                    (start_frame + idx) * sample_rate, image_format
                ),
            ),
            format=image_format,
            **image_kwargs,
        )
        for idx, img in enumerate(video_iterator)
    ]


video_metadata_schema = StructType(
    [
        StructField("width", IntegerType(), True),
        StructField("height", IntegerType(), True),
        StructField("num_frames", IntegerType(), True),
        StructField("duration", FloatType(), True),
        StructField("bit_rate", IntegerType(), True),
        StructField("frame_rate", IntegerType(), True),
        StructField("codec", StringType(), True),
        StructField("size", IntegerType(), True),
        StructField(
            "_errors",
            StructType(
                [
                    StructField("message", StringType(), True),
                    StructField("stderr", StringType(), True),
                    StructField("probe", StringType(), True),
                ]
            ),
            True,
        ),
    ]
)


@udf(returnType=video_metadata_schema)
def video_metadata(video: Union[str, VideoStream]) -> dict:
    """Return useful video stream metadata like width, height, num_frames,
    duration, bit_rate, frame_rate, codec, and size about the given video

    Parameters
    ----------
    video: str or VideoStream
        The video uri or the Video object

    Returns
    -------
    result: dict
        The keys are: width, height, num_frames, duration, bit_rate,
        frame_rate, codec, size, and _errors

    Notes
    -----
    The frame_rate is rounded to the nearest whole number of frames per sec

    Examples
    --------
    The following returns the fps rounded to the nearest integer:

    import rikai.spark.functions as RF
    (spark.createDataFrame([(VideoStream(<uri>),)], ['video'])
    .withColumn('meta', RF.video_metadata('video'))
    .select('meta.data.frame_rate'))

    """
    probe_result = _probe(video)
    if probe_result.get("_errors", None):
        return probe_result

    video_stream = next(
        (
            stream
            for stream in probe_result.get("streams", [])
            if stream.get("codec_type", None) == "video"
        ),
        None,
    )
    if video_stream is None:
        return _error("No video stream found", probe=probe_result)

    try:
        return {
            "width": _int_or_none(video_stream, "width"),
            "height": _int_or_none(video_stream, "height"),
            "num_frames": _int_or_none(video_stream, "nb_frames"),
            "frame_rate": _fps_or_none(video_stream),
            "duration": _float_or_none(video_stream, "duration"),
            "bit_rate": _int_or_none(video_stream, "bit_rate"),
            "codec": video_stream.get("codec_name", None),
            "size": _int_or_none(probe_result.get("format", {}), "size"),
        }
    except Exception as e:
        return _error(str(e))


def _probe(video):
    try:
        import ffmpeg
    except ImportError as e:
        raise ImportError(
            "Couldn't import ffmpeg. Please make sure to "
            "`pip install ffmpeg-python` explicitly or install "
            "the correct extras like `pip install rikai[video]`"
        ) from e

    uri = video.uri if isinstance(video, VideoStream) else video
    try:
        return ffmpeg.probe(uri)
    except Exception as e:
        return _error(str(e), stderr=getattr(e, "stderr", None))


def _int_or_none(data, key):
    return int(data[key]) if key in data else None


def _float_or_none(data, key):
    return float(data[key] if key in data else None)


def _fps_or_none(data: dict) -> Optional[int]:
    # ffprobe returns frame rate in timebase units (as a fraction)
    avg_frame_rate = data.get("avg_frame_rate", None)
    if avg_frame_rate is None:
        return None
    n, d = [int(x) for x in avg_frame_rate.split("/")]
    if d == 0:
        raise ValueError("avg_frame_rate had 0 time base")
    return int(round(n / d))


def _error(message, stderr=None, probe=None):
    err = {"message": message}
    if stderr is not None:
        err["stderr"] = stderr.decode("utf-8", "backslashreplace")
    if probe is not None:
        err["probe"] = str(probe)
    return {"_errors": err}


@udf(returnType=ImageType())
def spectrogram_image(
    video: Union[VideoStream, YouTubeVideo],
    output_uri: str,
    segment: Segment = Segment(0, -1),
    size: int = 224,
    max_samples: int = 15000,
    image_format: str = None,
    **image_kwargs,
) -> Image:
    """Applies ffmpeg filter to generate spectrogram image.

    Parameters
    ----------
    video : VideoStream or YouTubeVideo
        A video object whose audio track will be converted to spectrogram
    output_uri: str
        The uri to which the spectrogram image will be written to
    segment: Segment
            A Segment object, localizing video in time to (start_fno, end_fno)
    max_samples : Int
            Yield at most this many frames (-1 means no max)
    size : Int
        Sets resolution of frequency, time spectrogram image.
    image_format : str, optional
        The image format to save as. See
        `supported formats <https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.save>`_ for details.
    image_kwargs : dict, optional
        Optional arguments to pass to `PIL.Image.save <https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.save>`_.

    Return
    ------
    Image
        Return an Image of the audio spectrogram.
    """  # noqa: E501
    try:
        import ffmpeg
    except ImportError:
        raise ValueError(
            "Couldn't import ffmpeg. Please make sure to "
            "`pip install ffmpeg-python` explicitly or install "
            "the correct extras like `pip install rikai[video]`"
        )
    assert isinstance(
        video, (YouTubeVideo, VideoStream)
    ), "Input type must be YouTubeVideo or VideoStream"
    assert isinstance(segment, Segment), "Second input type must be Segment"

    start_frame = segment.start_fno
    if segment.end_fno > 0:
        max_samples = min((segment.end_fno - start_frame), max_samples)
    video_uri = (
        video.get_stream().uri
        if isinstance(video, YouTubeVideo)
        else video.uri
    )
    output, _ = (
        ffmpeg.input(video_uri)
        .filter("showspectrumpic", "{}x{}".format(size, size), legend=0)
        .output(
            "pipe:",
            format="rawvideo",
            pix_fmt="rgb24",
            start_number=start_frame,
            vframes=max_samples,
        )
        .run(capture_stdout=True)
    )
    return Image.from_array(
        np.frombuffer(output, np.uint8).reshape([size, size, 3]),
        output_uri,
        format=image_format,
        **image_kwargs,
    )


@udf(returnType=ArrayType(ImageType(), True))
def crop(img: Image, box: Union[Box2d, List[Box2d]]):
    """Crop image specified by the bounding boxes, and returns the cropped
    images.

    Parameters
    ----------
    img : Image
        An image object to be cropped.
    box : Box2d or List of Box2d
        One bound-2d box or a list of such boxes.

    Returns
    -------
    [Image]
        Return a list of cropped images.

    Examples
    --------

    >>> spark.sql("SELECT crop(img, boxes) as patches FROM detections")

    """
    return img.crop(box)



#  Copyright 2022 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Rikai-implemented PyTorch models and executors."""
from abc import ABC
from typing import Optional

import torch

from rikai.spark.sql.model import ModelSpec, ModelType

__all__ = ["TorchModelType"]


class TorchModelType(ModelType, ABC):
    """Base ModelType for PyTorch models."""

    def __init__(self):
        self.model: Optional[torch.nn.Module] = None
        self.spec: Optional[ModelSpec] = None

    def load_model(self, spec: ModelSpec, **kwargs):
        self.model = spec.load_model()
        self.model.eval()
        if "device" in kwargs:
            self.model.to(kwargs.get("device"))
        self.spec = spec

    # Release GPU memory
    # https://blog.paperspace.com/pytorch-memory-multi-gpu-debugging/?ref=tfrecipes
    def release(self):
        model = self.model.cpu()
        del model
        torch.cuda.empty_cache()



#  Copyright 2020 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Mixins
"""
from __future__ import annotations

from abc import ABC, abstractmethod
from io import BytesIO
from pathlib import Path
from typing import BinaryIO, Mapping, Optional, Union

# Third Party
import numpy as np

# Rikai
from rikai.internal.uri_utils import uri_equal
from rikai.io import open_uri

__all__ = ["ToNumpy", "Asset", "Displayable", "Drawable", "ToDict"]


class ToNumpy(ABC):
    """ToNumpy Mixin."""

    @abstractmethod
    def to_numpy(self) -> np.ndarray:
        """Returns the content as a numpy ndarray."""


class ToPIL(ABC):
    """ToPIL Mixin."""

    @abstractmethod
    def to_pil(self) -> "PIL.Image.Image":
        pass


class ToDict(ABC):
    """ToDict Mixin"""

    @abstractmethod
    def to_dict(self) -> dict:
        pass


class Displayable(ABC):
    """Mixin for notebook viz"""

    @abstractmethod
    def display(self, **kwargs) -> "IPython.display.DisplayObject":
        """Return an IPython.display.DisplayObject"""


class Drawable(ABC):
    """Mixin for a class that is drawable"""

    @abstractmethod
    def _render(self, render: "rikai.viz.Renderer", **kwargs) -> None:
        """Render the object using render."""

    def __matmul__(self, style: Union[dict, "rikai.viz.Style"]) -> Drawable:
        from rikai.viz import Style

        if not isinstance(style, (Mapping, Style)):
            raise ValueError(
                f"Must decorate drawable with a dict or style object"
                " but got {style}"
            )
        if isinstance(style, dict):
            style = Style(**style)
        return style(self)


class Asset(ABC):
    """cloud asset Mixin.

    Rikai uses asset to store certain blob on the cloud storage, to facilitate
    the functionality like fast query, example inspections, and etc.

    An asset is also a cell in a DataFrame for analytics. It offers both fast
    query on columnar format and easy tooling to access the actual data.

    Attributes
    ----------
    data : bytes, optional
        Embedded data
    uri : str
        URI of the external storage.
    """

    def __init__(
        self, data: Optional[bytes] = None, uri: Union[str, Path] = None
    ) -> None:
        assert (data is None) ^ (uri is None)
        assert data is None or isinstance(
            data, (bytes, bytearray)
        ), f"Expect bytes got {type(data)}"
        assert uri is None or isinstance(
            uri, (str, Path)
        ), f"Expect str/Path for uri, got {type(uri)}"
        self.data = data
        self.uri = str(uri) if uri is not None else None

    def __eq__(self, o: object) -> bool:
        return isinstance(o, Asset) and uri_equal(self.uri, o.uri)

    @property
    def is_embedded(self) -> bool:
        """Returns True if this Asset has embedded data."""
        return self.data is not None

    def open(self, mode="rb") -> BinaryIO:
        """Open the asset and returned as random-accessible file object."""
        if self.is_embedded:
            return BytesIO(self.data)
        return open_uri(self.uri, mode=mode)



import os
import pathlib

from setuptools import find_packages, setup

about = {}
with open(pathlib.Path("rikai") / "__version__.py", "r") as fh:
    exec(fh.read(), about)

with open(
    pathlib.Path(__file__).absolute().parent.parent / "README.md", "r"
) as fh:
    long_description = fh.read()

# extras
dev = [
    "black==22.1.0",
    "bump2version",
    "flake8",
    "isort",
    "pylint",
    "pytest-timeout",
    "pytest",
    "requests-mock",
    "wheel",
]
sklearn = ["scikit-learn"]
torch = ["torch>=1.8.1", "torchvision>=0.9.1"]
tf = ["tensorflow>=2.5.0", "tensorflow_hub"]
jupyter = ["matplotlib", "jupyterlab"]
aws = ["boto3", "botocore"]
gcp = ["gcsfs"]
docs = ["sphinx"]
video = ["ffmpeg-python", "scenedetect"]
youtube = ["pafy", "youtube_dl"]
mlflow = ["mlflow==1.22"]
all = (
    sklearn
    + torch
    + tf
    + jupyter
    + gcp
    + docs
    + video
    + youtube
    + mlflow
    + aws
)

if os.environ.get("SPARK_VERSION", None):
    spark_version = os.environ["SPARK_VERSION"]
else:
    spark_version = "3.1.2"

setup(
    name="rikai",
    version=about["version"],
    license="Apache License, Version 2.0",
    author="Rikai authors",
    author_email="rikai-dev@eto.ai",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/eto-ai/rikai",
    packages=find_packages(),
    include_package_data=True,
    python_requires=">=3.7",
    install_requires=[
        "antlr4-python3-runtime==4.8",
        "ipython>=7.31.1,!=8.0.0",
        "jsonschema",
        "numpy",
        "opencv-python-headless",
        "pandas",
        "Pillow",
        "pyarrow>=6.0",
        f"pyspark=={spark_version}",
        "pyyaml",
        "requests",
        "semver",
    ],
    extras_require={
        "all": all,
        "dev": dev,
        "docs": docs,
        "gcp": gcp,
        "jupyter": jupyter,
        "mlflow": mlflow,
        "sklearn": sklearn,
        "pytorch": torch,
        "tf": tf,
        "video": video,
        "youtube": youtube,
        "aws": aws,
    },
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Libraries",
    ],
)



#  Copyright 2022 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""ModelSpecs for official torchvision models
"""

from typing import Any, Callable

import torch.nn
from torchvision.transforms import ToTensor

from rikai.types import Box2d

from . import TorchModelType

__all__ = ["ObjectDetectionModelType"]


DEFAULT_MIN_SCORE = 0.5


class ObjectDetectionModelType(TorchModelType):
    """Shared ModelSpec for object detections in Torchvision

    https://pytorch.org/vision/stable/models.html
    """

    def __init__(self, name: str):
        super().__init__()
        self.name = name

    def __repr__(self):
        return f"ModelType({self.name})"

    def schema(self) -> str:
        return "array<struct<box:box2d, score:float, label_id:int>>"

    def transform(self) -> Callable:
        return ToTensor()

    def predict(self, images, *args, **kwargs) -> Any:
        assert (
            self.model is not None
        ), "model has not been initialized via load_model"
        min_score = float(
            self.spec.options.get("min_score", DEFAULT_MIN_SCORE)
        )

        batch = self.model(images)
        results = []
        for predicts in batch:
            predict_result = []
            for box, label, score in zip(
                predicts["boxes"].tolist(),
                predicts["labels"].tolist(),
                predicts["scores"].tolist(),
            ):
                if score < min_score:
                    continue
                predict_result.append(
                    {
                        "box": Box2d(*box),
                        "label_id": label,
                        "score": score,
                    }
                )

            results.append(predict_result)
        return results
